// -*- C++ -*-
// $RCSfile: oofcanvas.C,v $
// $Revision: 1.89 $
// $Author: langer $
// $Date: 2004/08/30 21:10:43 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/IO/stringimage.h"
#include "common/geometry.h"
#include "common/ooferror.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "common/IO/GUI/oofcanvas.h"
#include "common/IO/GUI/rubberband.h"
#include "common/IO/GUI/canvasdot.h"
#include "common/IO/GUI/canvastriangle.h"
#include "common/IO/threadstate.h"
#include <gdk_imlib.h>
#include <iostream>
#include <pygtk/pygtk.h>
#include <pygtk/pygdkimlib.h>

// Special for draw_seg_with_caps
#include <gdk/gdktypes.h>

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

bool OOFCanvas::initialized = 0;

OOFCanvas::OOFCanvas(bool antialias)
  : pixels_per_unit(1.0),
    current_layer(0),
    lineWidth(0),
    lineColor(0),
    fillColor(0),
    rubberband(0),
    isnew(true),
    empty(true),
    mousedown(false),
    mouse_callback(0),
    config_callback(0),
    mouse_handler_id(0),
    config_handler_id(0),
    underlayer(0),
    margin(1.0)
{
  if(!OOFCanvas::initialized) {
    OOFCanvas::initialized = 1;
    gtk_canvas_init();
    gdk_imlib_init();
    
    // The init_pygtk() macro in pygtk.h needs an extra cast to make
    // the C++ compiler happy, so we have to explicitly include a
    // modified version of the macro.
    acquirePyLock();
    PyObject *pygtk = PyImport_ImportModule("_gtk");
    if(pygtk != NULL) {
      PyObject *module_dict = PyModule_GetDict(pygtk);
      PyObject *cobject = PyDict_GetItemString(module_dict, "_PyGtk_API");
      if(PyCObject_Check(cobject))
	// extra cast required here
	_PyGtk_API = (_PyGtk_FunctionStruct*) PyCObject_AsVoidPtr(cobject);
      else {
	Py_FatalError("OOFCanvas: could not find _PyGtk_API object");
	releasePyLock();
	return;
      }
    }
    else {
      Py_FatalError("OOFCanvas: could not import _gtk");
      releasePyLock();
      return;
    }
    // end of init_pygtk
    releasePyLock();
  }

  if(antialias) {
    gtk_widget_push_visual(gdk_rgb_get_visual());
    gtk_widget_push_colormap(gdk_rgb_get_cmap());
    canvas = gtk_canvas_new_aa();
    gtk_widget_pop_colormap();
    gtk_widget_pop_visual();
  }
  else {
    gtk_widget_push_visual(gdk_imlib_get_visual());
    gtk_widget_push_colormap(gdk_imlib_get_colormap());
    canvas = gtk_canvas_new();
    gtk_widget_pop_colormap();
    gtk_widget_pop_visual();
  }

  // Don't get too many motion events.  This *should* mean that we
  // have to explicitly call gdk_window_get_pointer, but for some
  // mysterious reason it isn't necessary.
  gtk_widget_add_events(canvas, GDK_POINTER_MOTION_HINT_MASK);

  root = gtk_canvas_root(GTK_CANVAS(canvas));

  gtk_signal_connect(GTK_OBJECT(canvas), "destroy",
		     GTK_SIGNAL_FUNC(OOFCanvas::gtk_destroy),
		     this);

  // Create the underlayer.
  underlayer = new OOFCanvasLayer(this);
  
  // Flip the y axis.  The default GTK+ coordinate system has y=0 at
  // the top of the window, which bothers us.  We can't apply the
  // transformation to the root of the canvas item tree, though,
  // because images can't be displayed with affine transformations.
  // So here we just define the transformation, and apply it later.
  art_affine_scale(affineflip, 1.0, -1.0);

  clear();
  show();
}


void OOFCanvas::set_underlay_params(const std::vector<gushort> *color,
				    double f) {
  set_margin_local(f);
  set_bgColor_local(color);
}

void OOFCanvas::set_bgColor(const std::vector<gushort> *color) {
  set_bgColor_local(color);
  underlay();
}

void OOFCanvas::set_margin(double f) {
  set_margin_local(f);
  underlay();
}

void OOFCanvas::set_bgColor_local(const std::vector<gushort> *color) {
  GtkStyle *style = gtk_style_copy(gtk_widget_get_style(canvas));
  GdkColor &bgcolor = style->bg[0];
  GdkColormap *colormap = gtk_widget_get_colormap(canvas);
  bgcolor.red = (*color)[0];
  bgcolor.green = (*color)[1];
  bgcolor.blue = (*color)[2];
  gdk_colormap_alloc_color(colormap, &style->bg[0], FALSE, TRUE);
  gtk_widget_set_style(canvas, style);
}

void OOFCanvas::set_margin_local(double f) {
  margin = f;
}

ScrolledOOFCanvas::ScrolledOOFCanvas(bool aa)
  : OOFCanvas(aa)
{
  // Type casts required by the SGI compiler.
  scrollwindow = gtk_scrolled_window_new((GtkAdjustment*) NULL,
					 (GtkAdjustment*) NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwindow),
				 GTK_POLICY_AUTOMATIC,
				 GTK_POLICY_AUTOMATIC);
  gtk_signal_connect(GTK_OBJECT(scrollwindow), "destroy",
		     GTK_SIGNAL_FUNC(ScrolledOOFCanvas::gtk_destroy),
		     this);
  gtk_container_add(GTK_CONTAINER(scrollwindow), canvas);
  gtk_widget_show(scrollwindow);
}

OOFCanvas::~OOFCanvas() {
  if(canvas && GTK_IS_OBJECT(canvas) && !GTK_OBJECT_DESTROYED(canvas)) {
    gtk_object_destroy(GTK_OBJECT(canvas));
  }
}

ScrolledOOFCanvas::~ScrolledOOFCanvas() {
  if(scrollwindow && GTK_IS_OBJECT(scrollwindow)
     && !GTK_OBJECT_DESTROYED(scrollwindow)) {
    gtk_object_destroy(GTK_OBJECT(scrollwindow));
  }
}

void OOFCanvas::add_self(PyObject *pygtk_container) {
  gtk_container_add(GTK_CONTAINER(PyGtk_Get(pygtk_container)), widget());
}

void OOFCanvas::attach_self_to_table(PyObject *pygtk_table,
				     guint *left_attach, guint *right_attach,
				     guint *top_attach, guint *bottom_attach,
				     GtkAttachOptions *xoptions,
				     GtkAttachOptions *yoptions,
				     guint *xpadding, guint *ypadding)
{
  GtkTable *table = GTK_TABLE(PyGtk_Get(pygtk_table));
  gtk_table_attach(table, widget(), 
		   *left_attach, *right_attach, *top_attach, *bottom_attach,
		   *xoptions, *yoptions, *xpadding, *ypadding);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// When the gtk canvas or scrollwindow is destroyed via a propagated
// gtk event, we *don't* want to delete the C++ OOFCanvas object. If
// destroying a gtk object were to call the destructor of the
// corresponding C++ object, then when a gtk scrollwindow is
// destroyed, the underlying OOFCanvas's destructor would be called
// twice: once when the gtk canvas's "destroy" callback called it, and
// once when the ScrolledOOFCanvas's destructor called it.  SO, when a
// gtk object is destroyed, the C++ object is notified of the fact,
// but it isn't destructed.

void OOFCanvas::gtk_destroy(GtkWidget*, gpointer data) {
//  Trace("OOFCanvas::gtk_destroy");
  OOFCanvas *oofcanvas = (OOFCanvas*)(data);
  oofcanvas->destroyed();
}

void ScrolledOOFCanvas::gtk_destroy(GtkWidget*, gpointer data) {
//  Trace("ScrolledOOFCanvas::gtk_destroy");
}

void OOFCanvas::destroyed() {
//  Trace("OOFCanvas::destroyed");
  for(int i=0; i<layers.size(); i++)
    delete layers[i];
  delete underlayer;
  
  if(mouse_callback) {
    gtk_signal_disconnect(GTK_OBJECT(root), mouse_handler_id);
    acquirePyLock();
    Py_XDECREF(mouse_callback);
    releasePyLock();
  }

  if(config_callback) {
    gtk_signal_disconnect(GTK_OBJECT(canvas), config_handler_id);
    acquirePyLock();
    Py_XDECREF(config_callback);
    releasePyLock();
  }

  canvas = 0;
}

void ScrolledOOFCanvas::destroyed() {
  OOFCanvas::destroyed();
  scrollwindow = 0;
}

void OOFCanvas::show() {
  if(!canvas) throw ErrNoCanvas();
  gtk_widget_show_all(canvas);
  underlay();
}

void OOFCanvas::close() {
  if(!canvas) throw ErrNoCanvas();
  gtk_widget_destroy(widget());
  destroyed();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Dimensions of the window, in pixels.

int OOFCanvas::get_width_in_pixels() const {
  if(!canvas) throw ErrNoCanvas();
  GtkAllocation &allocation = canvas->allocation;
  return allocation.width;
}

int OOFCanvas::get_height_in_pixels() const {
  if(!canvas) throw ErrNoCanvas();
  GtkAllocation &allocation = canvas->allocation;
  return allocation.height;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Limits of everything that's been drawn, independent of the scroll region.

CRectangle OOFCanvas::get_bounds() const {
  if(!canvas) throw ErrNoCanvas();
  double x1, x2, y1, y2;
  gtk_canvas_item_get_bounds(GTK_CANVAS_ITEM(root), &x1, &y1, &x2, &y2);
  return CRectangle(Coord(x1, y1), Coord(x2, y2));
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Scaling functions

void OOFCanvas::set_pixels_per_unit(double ppu) {
  if(!canvas) throw ErrNoCanvas();
  pixels_per_unit = ppu;
  gtk_canvas_set_pixels_per_unit(GTK_CANVAS(canvas), ppu);
}

void OOFCanvas::zoom(double zoomfactor) {
  if(!canvas) throw ErrNoCanvas();
  pixels_per_unit *= zoomfactor;
  gtk_canvas_set_pixels_per_unit(GTK_CANVAS(canvas), pixels_per_unit);
  underlay();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void OOFCanvas::set_scrollregion(const CRectangle &rect) {
  if(!canvas) throw ErrNoCanvas();
  gtk_canvas_set_scroll_region(GTK_CANVAS(canvas),
			       rect.xmin(), rect.ymin(),
			       rect.xmax(), rect.ymax());
}

CRectangle OOFCanvas::get_scrollregion() const {
  if(!canvas) throw ErrNoCanvas();
  double x1, y1, x2, y2;
  gtk_canvas_get_scroll_region(GTK_CANVAS(canvas), &x1, &y1, &x2, &y2);
  return CRectangle(Coord(x1, y1), Coord(x2, y2));
}

void OOFCanvas::set_hadjustment(PyObject *adj) {
  gtk_layout_set_hadjustment(GTK_LAYOUT(canvas),
			     GTK_ADJUSTMENT(((PyGtk_Object*) adj)->obj));
  hadj = gtk_layout_get_hadjustment(GTK_LAYOUT(canvas));
}

void OOFCanvas::set_vadjustment(PyObject *adj) {
  gtk_layout_set_vadjustment(GTK_LAYOUT(canvas),
			     GTK_ADJUSTMENT(((PyGtk_Object*) adj)->obj));
  vadj = gtk_layout_get_vadjustment(GTK_LAYOUT(canvas));
}

Coord OOFCanvas::world_coord(int cx, int cy) const {
  double x, y;
  gtk_canvas_window_to_world(GTK_CANVAS(canvas), cx, cy, &x, &y);
  return Coord(x, y);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// "Underlay" function -- draws a rectangle in the background color
// which can catch mouse events outside the image.
void OOFCanvas::underlay() {
  // Remove the old underlay object.
  underlayer->clear();

  double x1, x2, y1, y2;
  // Get the bounds in the absence of the old underlayer.
  gtk_canvas_item_get_bounds(GTK_CANVAS_ITEM(root), &x1, &y1, &x2, &y2);
  double xmargin = (x2-x1)*margin;
  double ymargin = (y2-y1)*margin;
  x1 -= xmargin; x2 += xmargin;
  y1 -= ymargin; y2 += ymargin;

  // Get the color.
  GtkStyle *style = gtk_style_copy(gtk_widget_get_style(canvas));
  GdkColor bgcolor = style->bg[0];
  
  // Draw yourself.
  GtkCanvasItem *item = gtk_canvas_item_new(underlayer->group(),
 					    gtk_canvas_rect_get_type(),
 					    "x1", x1,
 					    "y1", y1,
 					    "x2", x2, 
 					    "y2", y2,
 					    "fill_color_gdk", &bgcolor,
 					    NULL);
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//


void OOFCanvas::clear() {
  CRectangle scrollregion = get_scrollregion();
  for(int i=0; i<layers.size(); i++) {
    layers[i]->destroy();
    delete layers[i];
  }
  GtkCanvasGroup *r = gtk_canvas_root(GTK_CANVAS(canvas));
  layers.resize(0);
  (void) newLayer();		// a new blank slate, so to speak
  set_scrollregion(scrollregion);
  empty = true;
}

void OOFCanvas::deleteLayer(OOFCanvasLayer *layer) {
  layer->destroy();
  for(std::vector<OOFCanvasLayer*>::iterator i=layers.begin(); i!=layers.end(); 
      ++i)
    {
      if(*i == layer) {
	layers.erase(i);
	return;
      }
    }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void OOFCanvas::set_lineWidth(unsigned int w) {
  lineWidth = w;
}

void OOFCanvas::set_lineColor(guint32 *color) {
  lineColor = *color;
}

void OOFCanvas::set_fillColor(guint32 *color) {
  fillColor = *color;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

OOFCanvasLayer *OOFCanvas::newLayer() {
  OOFCanvasLayer *layer = new OOFCanvasLayer(this);
  layers.push_back(layer);
  current_layer = layer;
  return layer;
}

OOFCanvasLayer::OOFCanvasLayer(OOFCanvas *canvas)
  : canvas(canvas)
{
  // By defining two groups, one within the other, we can clear a
  // layer without changing its z-position in the canvas.  The outer
  // group (grp) maintains the position, while the inner group
  // (subgrp) can be destroyed and recreated.
  grp = gtk_canvas_item_new(canvas->root,
			    gtk_canvas_group_get_type(),
			    "x", 0.0,
			    "y", 0.0,
			    NULL);
  subgrp = gtk_canvas_item_new(GTK_CANVAS_GROUP(grp),
			       gtk_canvas_group_get_type(),
			       "x", 0.0,
			       "y", 0.0,
			       NULL);
}

OOFCanvasLayer::~OOFCanvasLayer() {
  // The gtk object will be destroyed when the canvas is destroyed, or
  // when OOFCanvasLayer::destroy() is called.  There's really nothing
  // to do here.  Move along, now.
}

void OOFCanvasLayer::make_current() {
  if(!canvas) throw ErrNoCanvas();
  canvas->current_layer = this;
}

void OOFCanvasLayer::clear() {
  // Destroy the subgroup and recreate an empty one.
  gtk_object_destroy(GTK_OBJECT(subgrp));
  subgrp = gtk_canvas_item_new(GTK_CANVAS_GROUP(grp),
			       gtk_canvas_group_get_type(),
			       "x", 0.0,
			       "y", 0.0,
			       NULL);
}

void OOFCanvasLayer::destroy() {
  gtk_object_destroy(GTK_OBJECT(grp));
  grp = 0;
  subgrp = 0;
}

GtkCanvasGroup *OOFCanvasLayer::group() const {
  return GTK_CANVAS_GROUP(subgrp);
}

void OOFCanvasLayer::raise_layer(int howfar) {
  gtk_canvas_item_raise(grp, howfar);
}

void OOFCanvasLayer::raise_to_top() {
  gtk_canvas_item_raise_to_top(grp);
}

void OOFCanvasLayer::lower_layer(int howfar) {
  gtk_canvas_item_lower(grp, howfar);
}

void OOFCanvasLayer::show() {
  gtk_canvas_item_show(grp);
}

void OOFCanvasLayer::hide() {
  gtk_canvas_item_hide(grp);
}

void OOFCanvasLayer::move(double dx, double dy) {
  gtk_canvas_item_move(grp, dx, dy);
}
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void OOFCanvas::draw_dot(const Coord *dot) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					    gtk_canvas_dot_get_type(),
					    "x", (*dot)(0),
					    "y", (*dot)(1),
					    "size", lineWidth,
					    "fill_color_rgba", lineColor,
					    NULL);
  gtk_canvas_item_affine_absolute(item, affineflip);
}

// Will eventually have to get the angle.
void OOFCanvas::draw_triangle(const Coord *triangle, double angle) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					    gtk_canvas_triangle_get_type(),
					    "x", (*triangle)(0),
					    "y", (*triangle)(1),
					    "size", lineWidth,
					    "angle", angle, 
					    "fill_color_rgba", lineColor,
					    NULL);
  gtk_canvas_item_affine_absolute(item, affineflip);
}

void OOFCanvas::draw_segment(const GtkCanvasPoints *points) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					    gtk_canvas_line_get_type(),
					    "points", points,
					    "fill_color_rgba", lineColor,
					    "width_pixels", lineWidth,
					    NULL);
  gtk_canvas_item_affine_absolute(item, affineflip);
}

void OOFCanvas::draw_segments(const std::vector<GtkCanvasPoints*> *segs) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  for(int i=0; i<segs->size(); i++)
    draw_segment((*segs)[i]);
}

void OOFCanvas::draw_curve(const GtkCanvasPoints *points) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					    gtk_canvas_line_get_type(),
					    "points", points,
					    "fill_color_rgba", lineColor,
					    "width_pixels", lineWidth,
					    NULL);
  gtk_canvas_item_affine_absolute(item, affineflip);
}

void OOFCanvas::draw_polygon(const GtkCanvasPoints *points) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  static bool first = 1;
  if (first) {
    std::cerr << "OOFCanvas::draw_polygon bbox=" << get_bounds() << std::endl;
    for(int i=0; i<points->num_points; i++)
      std::cerr << "   " << points->coords[2*i] << ", "
		<< points->coords[2*i+1] << std::endl;
  }
  GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					    gtk_canvas_polygon_get_type(),
					    "points", points,
					    "outline_color_rgba", lineColor,
					    "width_pixels", lineWidth,
					    NULL);
  gtk_canvas_item_affine_absolute(item, affineflip);
  if(first) {
    std::cerr << "OOFCanvas::draw_polygon bbox=" << get_bounds() << std::endl;
    first = 0;
  }
}

void OOFCanvas::fill_polygon(const GtkCanvasPoints *points) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					    gtk_canvas_polygon_get_type(),
					    "points", points,
					    "fill_color_rgba", fillColor,
					    NULL);
  gtk_canvas_item_affine_absolute(item, affineflip);
}


// Use the negative of the y-coordinates provided, because
// affine transformations fail on circles.  As a result of this,
// the y coordinates are also in a different order, because
// there is a requirement that y1 be less than y2.
void OOFCanvas::draw_circle(const Coord &center, double radius) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					    gtk_canvas_ellipse_get_type(),
					    "x1", center(0) - radius,
					    "y1", -center(1) - radius,
					    "x2", center(0) + radius,
					    "y2", -center(1) + radius,
					    "outline_color_rgba", lineColor,
					    NULL);
  // gtk_canvas_item_affine_absolute(item, affineflip);
}

// See comment for draw_circle wrt affine transformations.
void OOFCanvas::fill_circle(const Coord &center, double radius) {
  if(!canvas) throw ErrNoCanvas();
  empty = false;
  GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					    gtk_canvas_ellipse_get_type(),
					    "x1", center(0) - radius,
					    "y1", -center(1) - radius,
					    "x2", center(0) + radius,
					    "y2", -center(1) + radius,
					    "fill_color_rgba", fillColor,
					    NULL);
  // gtk_canvas_item_affine_absolute(item, affineflip);
}


void OOFCanvas::draw_image(const StringImage *stringimage,
			   int iw, int ih,
			   const Coord &location,
			   double width, double height)
{
  if(!canvas) throw ErrNoCanvas();
  if(iw > 0 && ih > 0) {
    empty = false;
    GdkImlibImage *image = 
      gdk_imlib_create_image_from_data(
	       const_cast<unsigned char*>(stringimage->getString()), 0, iw, ih);
    GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					      gtk_canvas_image_get_type(),
					      "image", image,
					      "x", location(0),
					      "y", location(1),
					      "anchor", GTK_ANCHOR_SOUTH_WEST,
					      "width", width,
					      "height", height,
					      NULL);
  }
}


void OOFCanvas::draw_shaped_image(const StringImage *stringimage,
				  int iw, int ih,
				  const Coord &location,
				  double width, double height,
				  GdkImlibColor *shape_color)
{
  if(!canvas) throw ErrNoCanvas();
  if(iw > 0 && ih > 0) {
    empty = false;
    GdkImlibImage *image = 
      gdk_imlib_create_image_from_data(
	       const_cast<unsigned char*>(stringimage->getString()), 0, iw, ih);
    gdk_imlib_set_image_shape(image, shape_color);
    GtkCanvasItem *item = gtk_canvas_item_new(current_layer->group(),
					      gtk_canvas_image_get_type(),
					      "image", image,
					      "x", location(0),
					      "y", location(1),
					      "anchor", GTK_ANCHOR_SOUTH_WEST,
					      "width", width,
					      "height", height,
					      NULL);
  }
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Mouse event handling

void OOFCanvas::set_mouse_callback(PyObject *callback) {
  // callback must be a Python callable object which takes four arguments:
  //    event:  a string, either 'up', 'down', or 'move'
  //        x:  x position of the event
  //        y:  y position of the event, measured going *up* the screen
  //        s:  an int, 0 = no shift key, 1 = shift key
  //        c:  an int, 0 = no control key, 1 = control key
  mouse_callback = callback;
  acquirePyLock();
  Py_XINCREF(callback);
  releasePyLock();
  mouse_handler_id = 
    gtk_signal_connect(GTK_OBJECT(root), "event",
		       GTK_SIGNAL_FUNC(OOFCanvas::mouse_event), this);
}

gint OOFCanvas::mouse_event(GtkCanvasItem *item, GdkEvent *event, gpointer data)
{
  OOFCanvas *oofcanvas = (OOFCanvas*)(data);
  oofcanvas->mouse_eventCB(item, event);
  return FALSE;   // Returning FALSE propagates the event to parent items.
}

void OOFCanvas::mouse_eventCB(GtkCanvasItem *item, GdkEvent *event) {
  PyObject *args = 0;
  bool shift;
  bool ctrl;
  // Steal the focus so that other widgets know that their turn is over.
  // gtk_widget_grab_focus(canvas);
  // Protect the Python interpreter calls from thread interference
  // by obtaining the global interpreter lock.
  acquirePyLock();
  try {
    switch(event->type) {
    case GDK_BUTTON_PRESS:
      shift = event->button.state & GDK_SHIFT_MASK;
      ctrl = event->button.state & GDK_CONTROL_MASK;
      args = Py_BuildValue("sddii", "down",
			   event->button.x, -event->button.y, // flip y axis!
			   shift, ctrl);
      mousedown_point = Coord(event->button.x, event->button.y);
      mousedown = true;
    //     rubberband = make_rubberband(Coord(event->button.x, event->button.y));
      break;
    case GDK_BUTTON_RELEASE:
      shift = event->button.state & GDK_SHIFT_MASK;
      ctrl = event->button.state & GDK_CONTROL_MASK;
      args = Py_BuildValue("sddii", "up",
			   event->button.x, -event->button.y, // flip y axis!
			   shift, ctrl);
      if(rubberband && rubberband->active()) {
	rubberband->stop();
      //       delete rubberband;
      //       rubberband = 0;
      }
      mousedown = false;
      break;
    case GDK_MOTION_NOTIFY:
      shift = event->button.state & GDK_SHIFT_MASK;
      ctrl = event->button.state & GDK_CONTROL_MASK;
      args = Py_BuildValue("sddii", "move",
			   event->motion.x, -event->motion.y, // flip y axis!
			   shift, ctrl);
      if(mousedown && rubberband) {
	if(!rubberband->active()) {
	  rubberband->start(mousedown_point);
	}
	rubberband->redraw(canvas, Coord(event->motion.x, event->motion.y));
      }
      break;
    }

    if(args) {
      // PyGtk_Block/Unblock call is required within the callback.
      // This is a GTK requirement not related to the Python global
      // interpreter lock.  GTK may have its own thread lock.
      PyGtk_BlockThreads();
      PyObject *result = PyObject_CallObject(mouse_callback, args);
      if (result == NULL) {
	if(PyGtk_FatalExceptions) {
	  gtk_main_quit();
	}
	PyErr_Print();
	PyErr_Clear();
      }
      PyGtk_UnblockThreads();	
      Py_XDECREF(args);
      Py_XDECREF(result);
    }
  }
  catch(...) {
    releasePyLock();// release the global interpreter lock
    throw;
  }
  releasePyLock();
}


// Configuration event handling -- mainly we want to catch resizes.

void OOFCanvas::set_configure_callback(PyObject *callback) {
  config_callback = callback;
  acquirePyLock();
  Py_XINCREF(callback);
  releasePyLock();
  config_handler_id = 
    gtk_signal_connect(GTK_OBJECT(canvas), "size-allocate", 
		       GTK_SIGNAL_FUNC(OOFCanvas::configure_event), this);
}

gint OOFCanvas::configure_event(GtkCanvas *item, GtkAllocation *alloc,
				gpointer data)
{
  OOFCanvas *local_canvas = (OOFCanvas*)data;
  local_canvas->configure_eventCB(alloc);
  return FALSE;
}

void OOFCanvas::configure_eventCB(GtkAllocation *a) {
  gtk_widget_grab_focus(canvas);
  acquirePyLock();
  PyObject *args = Py_BuildValue("iiii", 
				 (int)(a->x), (int)(a->y), 
				 (int)(a->width), (int)(a->height));
  if (args) {
    PyGtk_BlockThreads();
    PyObject *result = PyObject_CallObject(config_callback, args);
    if (result==NULL) {
      if (PyGtk_FatalExceptions) {
	gtk_main_quit();
      }
      PyErr_Print();
      PyErr_Clear();
    }
    PyGtk_UnblockThreads();
    Py_XDECREF(result);
    Py_XDECREF(args);
  }
  releasePyLock();
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void OOFCanvas::set_rubberband(RubberBand *rb) {
  rubberband = rb;
}
