# -*- python -*-
# $RCSfile: regclassfactory.py,v $
# $Revision: 1.27 $
# $Author: langer $
# $Date: 2004/07/30 19:14:20 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.


# The RegisteredClassFactory is the ParameterWidget associated with
# Parameters for RegisteredClass variables.  As such it is created
# automatically inside ParameterTables, ParameterDialogs, MenuDialogs,
# and other RegisteredClassFactories.  It can also be used directly in
# GUI pages.

from SWIG.common import ooferror
from SWIG.common import switchboard
from common import debug
from common import guitop
from common import registeredclass
from common.IO import parameter
from common.IO.GUI import gui_utils
from common.IO.GUI import parameterwidgets
from common.IO.GUI import widgetscope
from gtk import *

###################

## RegisteredClass Parameter classes that have specialized widgets for
## some or all of their subclasses can register those widgets by
## calling addWidget().

widgetdictdict = {}

def addWidget(parameterclass, registeredclass, widgetclass):
    try:
        widgetdict = widgetdictdict[parameterclass]
    except KeyError:
        widgetdict = widgetdictdict[parameterclass] = {}
    widgetdict[registeredclass] = widgetclass

def _getWidgetDict(param):
    try:
        return widgetdictdict[param.__class__]
    except KeyError:
        return {}
    
####################

class RCFBase(parameterwidgets.ParameterWidget,
              widgetscope.WidgetScope):
    def __init__(self, gtk, scope, widgetdict):
        parameterwidgets.ParameterWidget.__init__(self, gtk, scope)
        widgetscope.WidgetScope.__init__(self, scope)

        # Dictionary of classes of widget wrappers for RegisteredClass
        # types that don't want to use a simple ParameterTable
        self.widgetdict = widgetdict

    # includeRegistration() controls which Registrations in the
    # registry will be listed in the RegisteredClassFactory.  Derived
    # classes can redefine this function. The default behavior is to
    # exclude secret registrations.
    def includeRegistration(self, registration):
        return not registration.secret

    def set_callback(self, callback, *args, **kwargs):
        self.callback = callback
        self.callbackargs = args
        self.callbackkwargs = kwargs
    
    def cleanUp(self):
        del self.registry               # break possible circular references?
        parameterwidgets.ParameterWidget.cleanUp(self)
        self.destroyScope()

####################

class RegisteredClassFactory(RCFBase):
    def __init__(self, registry, obj=None, title=None,
                 callback=None, fill=0, expand=0, scope=None,
                 widgetdict={},
                 *args, **kwargs):

        debug.mainthreadTest()
        self.registry = registry

        # The optionally supplied callback is called when a new
        # subclass is selected.  The args are the subclass's
        # registration, plus the extra args and kwargs given to
        # __init__.
        self.callback = callback
        self.callbackargs = args
        self.callbackkwargs = kwargs

        self.fill = fill
        self.expand = expand    
        # fill & expand option passed in to the add command in
        # "setByRegistration" routine.

        self.itemdict = {}              # used in sensitize()
        
        gtk = GtkFrame()
        RCFBase.__init__(self, gtk, scope, widgetdict)
        
        self.box = GtkVBox()
        self.gtk.add(self.box)
        self.options = gui_utils.SizedOptionMenu()
      
        if not title:
            self.box.pack_start(self.options, expand=0, fill=0)
            self.titlebox = None
        else:
            self.titlebox=GtkHBox()
            self.box.pack_start(self.titlebox, expand=0, fill=0)
            self.titlebox.add(GtkLabel(title), expand=0, fill=0)
            self.titlebox.add(self.options, expand=1, fill=1)

        self.widgetcallback = None
        
        self.menu = None
        self.paramWidget = None
        self.update(registry, obj)

    def update(self, registry, obj=None):
        debug.mainthreadTest()
        self.registry = registry
        self.itemdict = {}
        if self.menu:
            self.options.remove_menu()
        self.menu = GtkMenu()

        for registration in registry:
            if self.includeRegistration(registration):
                menuitem = GtkMenuItem(registration.name())
                self.itemdict[registration.name()] = menuitem
                if registration.tip:
                    guitop.top().tooltips.set_tip(menuitem, registration.tip)
                self.menu.add(menuitem)
                menuitem.connect('activate', self.optionCB, registration)
        self.options.set_menu(self.menu)
        if self.paramWidget:
            switchboard.removeCallback(self.widgetcallback)
            self.paramWidget.destroy()
        self.paramWidget = None
        self.widgetcallback = None
        self.currentOption = None
        self.show()

        # Set initial values
        if obj is not None:
            self.set(obj)
        else:
            # set to first acceptable registration
            for registration in registry:
                if self.includeRegistration(registration):
                    self.setByRegistration(registration)
                    return

    def set(self, obj):
        registration = obj.getRegistration()
        if self.includeRegistration(registration):
            # Set the registration's parameters
            obj.setDefaultParams()
            self.setByRegistration(obj.getRegistration())

    # Returns the currently-selected registration, or None.
    def getRegistration(self):
        return self.currentOption

    def sensitize(self, name):
        debug.mainthreadTest()
        self.itemdict[name].set_sensitive(1)
    def desensitize(self, name):
        debug.mainthreadTest()
        self.itemdict[name].set_sensitive(0)
        
    def setByRegistration(self, registration):
        debug.mainthreadTest()
        # setByRegistration() is almost the same as optionCB(), except
        # that it makes no attempt to set the new parameters from the
        # old ones, since it's not being used to switch between
        # convertible classes.  In fact, it may be used to switch to a
        # new instance of the same class, so it's important to throw
        # out the old parameter values.
        if self.includeRegistration(registration):
            if self.paramWidget:
                self.paramWidget.destroy()
            # We need to find the given registration's position in the
            # option menu.  Since not all registrations are listed,
            # search the registry, counting the listed registrations.
            counter = 0
            for reg in self.registry:
                if reg is registration:
                    self.options.set_history(counter)
                    break
                if self.includeRegistration(reg):
                    counter += 1
                
            self.getBase(registration)

            self.currentOption = registration

            self.paramWidget = self.makeWidget(registration)

            self.box.pack_start(self.paramWidget.gtk, fill=self.fill,
                                expand=self.expand)
            self.show()
            if hasattr(registration, 'tip'):
                guitop.top().tooltips.set_tip(self.options, registration.tip)

##            if self.callback:
##                apply(self.callback, (registration,)+self.callbackargs,
##                      self.callbackkwargs)

    def cleanUp(self):
        RCFBase.cleanUp(self)
        switchboard.removeCallback(self.widgetcallback)

    def show(self):
        debug.mainthreadTest()
        if self.titlebox is not None:
            self.titlebox.show_all()
        if self.paramWidget is not None:
            self.paramWidget.show()
        self.options.show_all()
        self.box.show()
        self.gtk.show()

    # Required by setByRegistration, but probably shouldn't be.
    def getBase(self, registration):
        pass
    
    # Actually instance a widget.  Convertible does this differently.
    def makeWidget(self, registration):
        try:
            # Use the special widget from the widgetdict, if there is one
            widget = self.widgetdict[registration.subclass](registration.params,
                                                            scope=self)
        except KeyError:
            # Otherwise, just use a parameter table.
            widget = parameterwidgets.ParameterTable(registration.params,
                                                     scope=self)
        self.widgetcallback = switchboard.requestCallback(('validity', widget),
                                                          self.vcheck)
        self.setValidity(widget.isValid())
        return widget

    def vcheck(self, validity):
        # Switchboard callback for ('validity', widget)
        self.setValidity(validity)

    def dumpValidity(self):
        self.paramWidget.dumpValidity()
    
    # When called as an event callback for menu selection in the
    # GtkOptionMenu, self.currentOption is either the "outgoing"
    # registration entry, or None.  
    def optionCB(self, menuitem, registration):
        debug.mainthreadTest()
        oldbase = None
        if self.currentOption is not None:
            if self.paramWidget is not None:
                # If there is a current option and a current widget,
                # copy the values into the parameters, so that they
                # become the defaults the next time the current widget
                # is displayed.  Some parameters may not yet have
                # legal values.  That's ok, since they're not going to
                # be used right away.
                try:
                    self.paramWidget.get_values()
                except parameter.ParameterMismatch:
                    pass
                self.paramWidget.destroy()
            
        # Set the current option to the new registration, assign params.
        # In the convertible case, this is a separate registration
        # instance, of a related (by convertibility) type.
        self.currentOption = registration
        self.optionFinish(registration)
        
    def optionFinish(self, registration):
        debug.mainthreadTest()
        self.paramWidget = self.makeWidget(registration)
        
        self.box.pack_start(self.paramWidget.gtk, fill=0, expand=0)
        self.show()
        if hasattr(registration, 'tip'):
            guitop.top().tooltips.set_tip(self.options, registration.tip)

        if self.callback:
            self.callback(registration, *self.callbackargs,
                          **self.callbackkwargs)

    def get_value(self):
        if self.currentOption is None:
            return None
        self.paramWidget.get_values()
        return self.currentOption()     # instantiates RegisteredClass

    def set_defaults(self):
        # Read widgets and set values in Registration, w/out creating object
        if self.currentOption is not None:
            self.paramWidget.get_values()

def _RegisteredClass_makeWidget(self, scope=None):
    return RegisteredClassFactory(self.registry, self.value, scope=scope,
                                  widgetdict=_getWidgetDict(self))

parameter.RegisteredParameter.makeWidget = _RegisteredClass_makeWidget

#####################################################################

# Subclass of the registeredclassfactory for convertible registered
# classes.  This is required because the widgets for convertible
# classes have more state information -- they need to know the
# original pre-instantaion value, in base form, as well as the
# current state data stored in the params.
class ConvertibleRegisteredClassFactory(RegisteredClassFactory):
    def __init__(self, registry, obj=None, title=None,
                 callback=None, fill=0, expand=0, scope=None, widgetdict={},
                 *args, **kwargs):
        debug.mainthreadTest()
        self.base_value = None
        RegisteredClassFactory.__init__(self, registry, obj,
                                        title=title,
                                        callback=callback,
                                        fill=fill,
                                        expand=expand,
                                        scope=scope,
                                        widgetdict=widgetdict,
                                        *args, **kwargs)
        
    # The ConvertibleRCF's optionCB needs to extract the values from
    # the "outgoing" widgets in order to insert them, suitably
    # converted, into the "incoming" widgets.  Widgets whose
    # parameters have nontrivial ".value" retrievals (e.g. dependent
    # parts of ParameterGroups) should ensure that their resolver
    # does not throw spurious exceptions in this circumstance.
    def optionCB(self, menuitem, registration):
        if self.currentOption is not None:
            if self.paramWidget is not None:
                # If there is a current option and a current widget,
                # copy the values into the parameters.
                self.paramWidget.get_values()
                self.paramWidget.destroy()
            # Retrieve the current value of the outgoing option.
            old = self.getParamValues()
            
        # Set the current option to the new registration, assign params.
        # In the convertible case, this is a separate registration
        # instance, of a related (by convertibility) type.
        self.currentOption = registration
        self.setParams(old)

        self.optionFinish(registration)

    # All the functions here override virtual functions in the
    # immediate base class.
    def getBase(self, registration):
        # Get the original value of the parameter, in base form,
        # and store it so you can pass it to widgets for other representations.
        self.base_value = registration.getParamValuesAsBase()

    def getParamValues(self):
        return self.currentOption.getParamValuesAsBase()

    def setParams(self, old):
        self.currentOption.setParamsFromBase(old)
        
    # In the convertible case, the widget constructor will take another
    # argument, the "base_value" which is the original value of the
    # parameter, in base form.  
    def makeWidget(self, registration):
        try:
            widget = self.widgetdict[registration.subclass](registration.params,
                                                            self.base_value,
                                                            scope=self)
        except KeyError:
            widget = parameterwidgets.ParameterTable(registration.params,
                                                     scope=self)
        self.widgetcallback = switchboard.requestCallback(('validity', widget),
                                                          self.vcheck)
        self.setValidity(widget.isValid())
        return widget
    
def _ConvertibleRegisteredClass_makeWidget(self, scope=None):
    return ConvertibleRegisteredClassFactory(self.registry, self.value,
                                             scope=scope,
                                             widgetdict=_getWidgetDict(self))

parameter.ConvertibleRegisteredParameter.makeWidget = \
                      _ConvertibleRegisteredClass_makeWidget


#####################################################################

# RegisteredClassListFactory

ParameterTable = parameterwidgets.ParameterTable

class RegistrationGUIData:
    def __init__(self, registration, rclfactory):
        debug.mainthreadTest()
        self.registration = registration
        self.rclfactory = rclfactory
        self._button = GtkCheckButton()
        self._signal = self._button.connect('clicked', self.buttonCB)
        self._box = GtkVBox()
        label = GtkLabel(registration.name())
        label.set_alignment(0.0, 0.5)
        if hasattr(registration, 'tip'):
            self._label = GtkEventBox()
            guitop.top().tooltips.set_tip(self._label, registration.tip)
            self._label.add(label)
        else:
            self._label = label
        self.sbcallback = None
        self.makeWidget()
    def makeWidget(self):
        debug.mainthreadTest()
        if self.sbcallback:
            switchboard.removeCallback(self.sbcallback)
        self._box.foreach(GtkObject.destroy)
        try:
            self._widget = self.rclfactory. \
                           widgetdict[self.registration.subclass] \
                           (registration.params, scope=self.rclfactory)
        except KeyError:                # no special widget defined
            self._widget = ParameterTable(self.registration.params,
                                          scope=self.rclfactory)
        self._box.add(self._widget.gtk)
        self.widgetcallback = switchboard.requestCallback(
            ('validity', self._widget), self.validityCB)
    def validityCB(self, validity):
        self.rclfactory.validityCheck()
    def isValid(self):
        return self._widget.isValid()
    def showWidget(self):
        debug.mainthreadTest()
        self._box.set_sensitive(self._button.get_active())
        self._box.show()
        self._widget.show()
    def name(self):
        return self.registration.name()
    def box(self):
        return self._box
    def button(self):
        return self._button
    def label(self):
        return self._label
    def setButton(self, active):
        debug.mainthreadTest()
        self._button.signal_handler_block(self._signal)
        self._button.set_active(active)
        self._button.signal_handler_unblock(self._signal)
        self.showWidget()
    def getButton(self):
        debug.mainthreadTest()
        return self._button.get_active()
    def buttonCB(self, *args):
        self.rclfactory.validityCheck()
        self.showWidget()
    def show(self):
        debug.mainthreadTest()
        self._label.show_all()
        self._button.show()
        self.showWidget()
    def get_value(self):
        debug.mainthreadTest()
        if self._button.get_active():
            self._widget.get_values()
            return self.registration()  # creates RegisteredClass instance
    def cleanUp(self):
        switchboard.removeCallback(self.widgetcallback)
        

class RegisteredClassListFactory(RCFBase):
    def __init__(self, registry, objlist=None, title=None, callback=None,
                 fill=0, expand=0, scope=None, widgetdict={},
                 *args, **kwargs):
        debug.mainthreadTest()

        self.registry = registry
        self.callback = callback
        self.callbackargs = args
        self.callbackkwargs = kwargs
        self.title = title
        self.fill = fill
        self.expand = expand

        gtk = GtkFrame()
        self.table = GtkTable(rows=1, cols=2)
        gtk.add(self.table)
        RCFBase.__init__(self, gtk, scope, widgetdict)
        self.parent.addWidget(self)
        self.update(registry, objlist)

    def cleanUp(self):
        del self.guidata
        RCFBase.cleanUp(self)

    def update(self, registry, objlist=[]):
        # Called during initialization and whenever the registry changes.
        debug.mainthreadTest()
        self.registry = registry
        self.table.foreach(GtkObject.destroy) # clear the table
        row = 0
        if self.title:
            self.table.resize(rows=2, cols=2)
            self.table.attach(GtkLabel(self.title), 0,2, row,row+1,
                              yoptions=0, ypadding=5)
            row += 1
            self.table.attach(GtkHSeparator(), 0,2, row,row+1,
                              yoptions=0, ypadding=3)
            row += 1
        else:                           # no title
            self.table.resize(rows=0, cols=2)
        self.guidata = {}               # RegistrationGUIData, keyed by reg name
        for registration in registry:
            if self.includeRegistration(registration):
                self.table.resize(rows=row+3, cols=2)
                data = RegistrationGUIData(registration, self)
                self.guidata[registration.name()] = data
                self.table.attach(data.button(), 0,1, row,row+1, xoptions=0,
                                  yoptions=0, xpadding=3, ypadding=3)
                self.table.attach(data.label(), 1,2, row,row+1,
                                  xoptions=EXPAND|FILL, yoptions=0)
                self.table.attach(data.box(), 1,2, row+1,row+2,
                                  xoptions=EXPAND|FILL, yoptions=0,
                                  xpadding=3, ypadding=3)
                self.table.attach(GtkHSeparator(), 0,2, row+2, row+3,
                                  xoptions=EXPAND|FILL, yoptions=0)
                row += 3
        self.table.resize(rows=row-1, cols=2) # remove last GtkHSeparator
        if objlist is not None:
            self.set(*objlist)
        self.validityCheck()

    def set(self, *objlist):
        for obj in objlist:
            obj.setDefaultParams()
        self.setByRegistrations([obj.getRegistration() for obj in objlist])

    def setByRegistrations(self, registrations):
        for reg in self.registry:
            if self.includeRegistration(reg):
                data = self.guidata[reg.name()]
                data.setButton(reg in registrations) # calls data.makeWidget
        self.show()

    def validityCheck(self):
        ok = 0
        for data in self.guidata.values():
            if data.getButton():
                if data.isValid():
                    ok = 1
                else:
                    self.setValidity(0)
                    return
        self.setValidity(ok)
            
    def show(self):
        debug.mainthreadTest()
        # Don't use self.gtk.show_all(), since it will show collapsed items.
        for data in self.guidata.values():
            data.show()
        self.table.show()
        self.gtk.show()
        
    def get_value(self):
        values = []
        for registration in self.registry:
            data = self.guidata[registration.name()]
            value = data.get_value()
            if value is not None:
                values.append(value)
        return values

    def set_defaults(self):
        # For the regular RegisteredClassFactory, this copies the
        # values out of the gui and into the Parameters in the
        # Registrations (which are the same as the Parameters in the
        # menus) so that calling menuitem.callWithDefaults() does the
        # right thing.  This doesn't make sense for the
        # RegisteredClassListFactory, since the menu Parameter is a
        # list of RegisteredClass objects, not the parameters of the
        # objects themselves.  So don't use this, use get_value() instead.
        raise ooferror.ErrPyProgrammingError(
            "Don't use RegisteredClassListFactory.set_defaults()!")

def _RegisteredClassList_makeWidget(self, scope=None):
    return RegisteredClassListFactory(self.registry, self.value, scope=scope,
                                      widgetdict=_getWidgetDict(self))

parameter.RegisteredListParameter.makeWidget = _RegisteredClassList_makeWidget
