// -*- C++ -*-
// $RCSfile: coord.h,v $
// $Revision: 1.16 $
// $Author: langer $
// $Date: 2004/04/20 18:29:14 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifndef COORD_H
#define COORD_H

class Coord;
class ICoord;
class Position;

#include "space.h"
#include <iostream>

// will have to be rewritten if spacedim changes!


// The Coord and MasterCoord classes used to be both derived from the
// same template, so that they could share the same code but still be
// distinct classes. That doesn't work now that Coord is derived from
// Position, but MasterCoord isn't.  So now the classes have
// separate but equal definitions.

// The Position base class allows simple objects like Coord and
// complex objects like GaussPoint to be passed to the same functions,
// if all they need to know is the position of the object.

class Position {
public:
  virtual ~Position () {}
  virtual Coord position() const = 0;
};

class Coord : public Position {
protected:
  double x[spacedim];
public:
  virtual ~Coord() {}
  Coord() { x[0] = x[1] = 0; }
  Coord(double x0, double x1) { x[0] = x0; x[1] = x1; }
  double operator()(int i) const { return x[i]; }
  double &operator()(int i) { return x[i]; }
  Coord &operator+=(const Coord &c) {
    x[0] += c(0);
    x[1] += c(1);
    return *this;
  }
  Coord &operator-=(const Coord &c) {
    x[0] -= c(0);
    x[1] -= c(1);
    return *this;
  }
  Coord &operator+=(const ICoord&);
  Coord &operator-=(const ICoord&);
  Coord &operator*=(double y) {
    x[0] *= y;
    x[1] *= y;
    return *this;
  }
  double normsq() const { return x[0]*x[0] + x[1]*x[1]; }
  friend bool operator==(const Coord&, const Coord&);
  friend bool operator!=(const Coord&, const Coord&);
  virtual Coord position() const { return *this; }
  virtual void copyCoord(const Coord &c) { x[0] = c.x[0]; x[1] = c.x[1]; }
};

inline bool operator<(const Coord &a, const Coord &b) {
  if(a(0) < b(0)) return true;
  if(a(0) > b(0)) return false;
  if(a(1) < b(1)) return true;
  return false;
}

std::ostream &operator<<(std::ostream&, const Coord&);

std::istream &operator>>(std::istream&, Coord&);

inline Coord operator+(const Coord &a, const Coord &b) {
  Coord result(a);
  result += b;
  return result;
}

inline Coord operator-(const Coord &a, const Coord &b) {
  Coord result(a);
  result -= b;
  return result;
}

inline Coord operator*(const Coord &a, double x) {
  Coord b(a);
  b *= x;
  return b;
}

inline Coord operator*(double x, const Coord &a) {
  Coord b(a);
  b *= x;
  return b;
}

inline double cross(const Coord &c1, const Coord &c2)
{
  return c1(0)*c2(1) - c1(1)*c2(0);
}

inline double operator%(const Coord &c1, const Coord &c2)
{
  return(cross(c1,c2));
}

inline bool operator==(const Coord &a, const Coord &b) {
  return a.x[0] == b.x[0] && a.x[1] == b.x[1];
}

inline bool operator!=(const Coord &a, const Coord &b) {
  return a.x[0] != b.x[0] || a.x[1] != b.x[1];
}

inline double dot(const Coord &c1, const Coord &c2) {
  return c1(0)*c2(0) + c1(1)*c2(1);
}

inline double norm2(const Coord &c) {
  return dot(c, c);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Integer coordinates

class ICoord {
protected:
  int x[spacedim];
public:
  virtual ~ICoord() {}
  ICoord() { x[0] = x[1] = 0; }
  ICoord(int x0, int x1) { x[0] = x0; x[1] = x1; }
  int operator()(int i) const { return x[i]; }
  int &operator()(int i) { return x[i]; }
  ICoord &operator+=(const ICoord &c) {
    x[0] += c(0);
    x[1] += c(1);
    return *this;
  }
  ICoord &operator-=(const ICoord &c) {
    x[0] -= c(0);
    x[1] -= c(1);
    return *this;
  }
  ICoord &operator*=(int y) {
    x[0] *= y;
    x[1] *= y;
    return *this;
  }
  friend bool operator==(const ICoord&, const ICoord&);
  friend bool operator!=(const ICoord&, const ICoord&);
};

std::ostream &operator<<(std::ostream&, const ICoord&);
std::istream &operator>>(std::istream&, ICoord&);

inline ICoord operator+(const ICoord &a, const ICoord &b) {
  ICoord result(a);
  result += b;
  return result;
}

inline ICoord operator-(const ICoord &a, const ICoord &b) {
  ICoord result(a);
  result -= b;
  return result;
}

inline Coord operator+(const Coord &a, const ICoord &b) {
  Coord result(a);
  result += b;
  return result;
}

inline Coord operator-(const Coord &a, const ICoord &b) {
  Coord result(a);
  result -= b;
  return result;
}

inline Coord operator+(const ICoord &b, const Coord &a) {
  Coord result(a);
  result += b;
  return result;
}

inline Coord operator-(const ICoord &b, const Coord &a) {
  Coord result(a);
  result -= b;
  return result;
}

inline ICoord operator*(const ICoord &a, int x) {
  ICoord b(a);
  b *= x;
  return b;
}

inline Coord operator*(const ICoord &a, double x) {
  Coord b(a(0), a(1));
  b *= x;
  return b;
}

inline bool operator==(const ICoord &a, const ICoord &b) {
  return a.x[0] == b.x[0] && a.x[1] == b.x[1];
}

inline bool operator!=(const ICoord &a, const ICoord &b) {
  return a.x[0] != b.x[0] || a.x[1] != b.x[1];
}

inline int dot(const ICoord &c1, const ICoord &c2) {
  return c1(0)*c2(0) + c1(1)*c2(1);
}

inline int norm2(const ICoord &c) {
  return dot(c, c);
}

bool operator<(const ICoord &a, const ICoord &b);

#endif // COORD_H
