# -*- python -*-
# $RCSfile: skeletonbdydisplay.py,v $
# $Revision: 1.10 $
# $Author: langer $
# $Date: 2004/09/16 19:14:18 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from common import color
from common import debug
from common import primitives
from common import registeredclass
from common.IO import display
from common.IO import gfxmanager
from common.IO import parameter
from engine import skeletonboundary
from SWIG.common import switchboard
import math
import string

# Special parameter class for skeleton boundaries.  Returns a
# list of strings corresponding to the names of boundaries in a
# given skeleton.  The custom class is primarily to allow for a
# special skeleton-smart widget.

class SkeletonBoundaryParameter(parameter.ListOfStringsParameter):
    pass

# Layer for showing the boundaries (point and edge) of a skeleton.
class SkeletonBoundaryDisplay(display.DisplayMethod):
    def __init__(self, boundaries, color, linewidth, dotsize, arrowsize):
        self.boundaries = boundaries
        self.color = color
        self.linewidth = linewidth
        self.dotsize = dotsize
        self.arrowsize = arrowsize
        display.DisplayMethod.__init__(self)

    def draw(self, gfxwindow, device):
        skel = self.who().resolve(gfxwindow)
        skelobj = skel.getObject()
        device.set_lineColor(self.color)
        device.set_lineWidth(self.linewidth)
        for k in self.boundaries:
            try:
                b = skelobj.edgeboundaries[k]
            except KeyError:
                pass
            else:
                for e in b.edges:
                    nodes = e.get_nodes()
                    device.set_lineWidth(self.linewidth)
                    device.draw_segment(primitives.Segment(
                        nodes[0].position(), nodes[1].position()))
                    # Boundaries are directed from 0 to 1.
                    center = (nodes[0].position() + nodes[1].position())/2
                    diff = (nodes[1].position() - nodes[0].position())
                    # Zero of angles is the y-axis, not the x-axis...
                    angle = math.atan2(-diff.x, diff.y)
                    device.set_lineWidth(self.arrowsize)
                    device.draw_triangle(center, angle)
            
        device.set_lineWidth(self.dotsize)
        device.set_fillColor(self.color)
        for k in self.boundaries:
            try:
                b = skelobj.pointboundaries[k]
            except KeyError:
                pass
            else:
                for n in b.nodes:
                    device.draw_dot(n.position())

    def getTimeStamp(self, gfxwindow):
        return max( self.timestamp,
                    self.who().resolve(gfxwindow).bdytimestamp )

    # Need to override hash, because we contain a list.  For this
    # class, object identity is a good test of equality.
    def __hash__(self):
        return id(self)
    
                    
skeletonBoundaryDisplay = registeredclass.Registration(
    'Boundaries', display.DisplayMethod,
    SkeletonBoundaryDisplay,
    params=[SkeletonBoundaryParameter('boundaries', []),
            color.ColorParameter('color', value=color.gray50),
            parameter.IntParameter('linewidth', 4),
            parameter.IntParameter('dotsize', 4),
            parameter.IntParameter('arrowsize', 10)],
    ordering=1.0,
    layerordering=display.SemiLinear(2),
    whoclasses=('Skeleton',),
    tip="Display some or all of the boundaries of the Skeleton.")

# Layer for showing the newest boundary (point and edge) of a skeleton.
class SelectedSkeletonBoundaryDisplay(display.DisplayMethod):
    def __init__(self, color, linewidth, dotsize, arrowsize):
        self.boundary = None
        self.color = color
        self.linewidth = linewidth
        self.dotsize = dotsize
        self.arrowsize = arrowsize
        display.DisplayMethod.__init__(self)
        self.sbcb = switchboard.requestCallback(
            "boundary selected", self.selBdyCB)

    def draw(self, gfxwindow, device):
        skel = self.who().resolve(gfxwindow)
        skelobj = skel.getObject()
        device.set_lineColor(self.color)
        device.set_lineWidth(self.linewidth)

        # Edge boundary
        try:
            b = skelobj.edgeboundaries[self.boundary]
        except KeyError:
            pass
        else:
            for e in b.edges:
                nodes = e.get_nodes()
                device.set_lineWidth(self.linewidth)
                device.draw_segment(primitives.Segment(
                    nodes[0].position(), nodes[1].position()))
                # Boundaries are directed from 0 to 1.
                center = (nodes[0].position() + nodes[1].position())/2
                diff = (nodes[1].position() - nodes[0].position())
                # Zero of angles is the y-axis, not the x-axis...
                angle = math.atan2(-diff.x, diff.y)
                device.set_lineWidth(self.arrowsize)
                device.draw_triangle(center, angle)

        # Point boundary
        device.set_lineWidth(self.dotsize)
        device.set_fillColor(self.color)
        try:
            b = skelobj.pointboundaries[self.boundary]
        except KeyError:
            pass
        else:
            for n in b.nodes:
                device.draw_dot(n.position())
    
    def selBdyCB(self, context, name):
        # Check if any gfxwindows displays the skeleton as a top layer.
        gfxwindows = gfxmanager.gfxManager.getAllWindows()
        drawable = 0
        for window in gfxwindows:
            skel = self.who().resolve(window)
            if context is skel:
                drawable = 1
                break
        if drawable:
            self.boundary = name

    def getTimeStamp(self, gfxwindow):
        return max( self.timestamp,
                    self.who().resolve(gfxwindow).bdytimestamp )

    # Need to override hash, because we contain a list.  For this
    # class, object identity is a good test of equality.
    def __hash__(self):
        return id(self)

    def __del__(self):
        switchboard.removeCallback(self.sbcb)
                    
selectedSkeletonBoundaryDisplay = registeredclass.Registration(
    'Selected Boundary', display.DisplayMethod,
    SelectedSkeletonBoundaryDisplay,
    params=[color.ColorParameter('color', value=color.orange),
            parameter.IntParameter('linewidth', 4),
            parameter.IntParameter('dotsize', 4),
            parameter.IntParameter('arrowsize', 10)],
    ordering=2.3,
    layerordering=display.SemiLinear(1),
    whoclasses=('Skeleton',),
    tip="Display the currently selected boundary.")
                                                

display.predefineLayer('Skeleton', '<topmost>',
                       selectedSkeletonBoundaryDisplay())
