# -*- python -*-
# $RCSfile: bdycondition.py,v $
# $Revision: 1.36 $
# $Author: reida $
# $Date: 2004/09/08 21:03:56 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# A new, leaner, meaner, boundary condition class.  For the
# point-like ones, the iteration over nodes and point evaluation,
# as required, will be done elsewhere.  All we need to be able
# to do here is take a node and a profile result, and apply it.

from SWIG.common import ooferror
from SWIG.common import switchboard
from SWIG.engine import boundarycond
from common import debug
from common import labeltree
from common import registeredclass
from common.IO import automatic
from common.IO import parameter
from common.IO import whoville
from engine import problem
from engine import profile
from engine import skeletoncontext
from engine.IO import meshparameters
from engine.IO import skeletongroupparams
import engine.mesh


def bcnameResolver(param, name):
    if param.automatic():
        basename = 'bc'
    else:
        basename = name
    meshname = param.group['mesh'].value
    if meshname is not None:
        meshpath = labeltree.makePath(meshname)
        meshcontext = engine.mesh.meshes[meshpath]
        return meshcontext.uniqueBCName(basename)



nameparam = parameter.AutomaticNameParameter(
    'name', value=automatic.automatic, tip="Select a name.",
    resolver=bcnameResolver)

meshparam = whoville.WhoParameter("mesh", engine.mesh.meshes,
                                  tip="Specify the mesh.")

bccontextgroup = parameter.ParameterGroup(
    nameparam,
    meshparam)




# It's useful in the editor if all the BC's have shared
# parameters, to the greatest degree possible.

bc_field_param_group = parameter.ParameterGroup(
    meshparameters.FieldParameter('field'),
    meshparameters.FieldIndexParameter('field_component'))

bc_equation_param_group = parameter.ParameterGroup(
    meshparameters.EquationParameter('equation'),
    meshparameters.FieldIndexParameter('eqn_component'))

bc_boundary_param = skeletongroupparams.BoundaryParameter(
    "boundary",tip="Boundary to which this condition applies.")



# Boundary conditions are constructed in association with a particular
# profile and boundary, and can't really exist without these.
# Boundary conditions cannot be shared between profiles or between
# boundaries.
class BC(registeredclass.RegisteredClass):
    registry = []
    def __init__(self, profile, boundary):

        self.profile = profile     # Profile object.
        self.boundary = boundary   # Name of the boundary.

        # Unpack the NamedProfile, if you got one.
        self.profile_obj = self.profile.unpack()

        self.mesh_obj = None

        self.sw_cb = [switchboard.requestCallback(
            "new namedprofile obj", self.replace_profile)]

    # Some boundary conditions are "secret" in the sense of not
    # appearing in the list on the GUI page.  The default is to not be
    # secret.
    def secret(self):
        return None

    def __eq__(self, other):
        return id(self)==id(other)
    
    # Add_to_mesh is separate from the construction process for the
    # convenience of the boundary-condition editing, which is
    # accomplished by creating a new boundary condition and adding it
    # to the mesh with the name of a recently-removed one. 
    def add_to_mesh(self, name, mesh, signal=1):
        self.name = name 
        self.mesh = mesh # Mesh name.
        
        self.mesh_obj = engine.mesh.meshes[mesh] # Who object.
        self.femesh = self.mesh_obj.femesh()     # Actual mesh.

        self.boundary_obj = self.mesh_obj.getBoundary(self.boundary)

        self.profile_obj.addCondition(self)
        self.boundary_obj.addCondition(self)
        self.mesh_obj.addBdyCondition(name, self, signal=signal)

    def remove_from_mesh(self, signal=1):
        self.mesh_obj.rmBdyCondition(self.name, signal=signal)
        # rmBdyCondition calls disconnect, below.


    # Called prior to deletion, undo the additions we did on __init__.
    # Because of the circular references, as well as control over the
    # timing, these operations are unsuitable for a __del__ operation.
    def disconnect(self):
        self.profile_obj.removeCondition(self)
        self.boundary_obj.removeCondition(self)
        # Break circular references.
        self.boundary = None
        self.boundary_obj = None
        self.profile=None
        self.profile_obj = None
        # Clean up callbacks.
        map(switchboard.removeCallback, self.sw_cb)


    # Called from the NamedProfile object when profiles are edited.
    # The NamedProfile has a new .obj, and the conditions must all
    # update accordingly.
    def replace_profile(self, profile):
        if self.profile == profile:
            self.profile_obj = self.profile.unpack()
            self.profile_obj.addCondition(self)

            
    def display_string(self):
        return self.boundary_obj.name() + " / " + self.display() + \
               " / %s" % self.profile_obj.descriptor()
    
##    def display_string(self):
##        bdyprofstring = " on boundary %s with profile %s." % \
##                        (`self.boundary_obj.name()`,
##                         self.profile_obj.descriptor())
##        return self.display() + bdyprofstring



# Individual boundary condition classes take the field and equation
# data in the form of a component object, which they must then change
# to a separate field and component for actual evaluation.  Boundary
# conditions are also specific to a particular mesh and boundary.
# They also know about a profile -- the object passed in may either be
# an existing profile in the current mesh, or a new unnamed profile.
# In either case, the passed in object is the appropriate subclass of
# ProfileFinder.

class DirichletBC(BC):
    def __init__(self, field, field_component, equation, eqn_component,
                 profile, boundary, 
                 offset=0.0, fromFloat = 0):
        self.field = field
        self.field_component = field_component
        self.equation = equation
        self.eqn_component = eqn_component

        # Parent class __init__ needs to be last, because it does
        # collision detection on the boundary, which means the field
        # and equation have to already be set.
        BC.__init__(self, profile, boundary)

        # When a DirichletBC is auto-generated from a FloatBC, it can
        # acquire an offset, and gets "fromFloat" set.  
        self.offset = offset
        self.fromFloat = fromFloat

    # Fixed BCs are secret if they're derived from a floating BC.
    def secret(self):
        return self.fromFloat
    
    # "Copy" makes a new BC with a new name on a new boundary, but
    # with the same other parameters.
    def copy(self, boundary):
        return DirichletBC(self.field, self.field_component,
                       self.equation, self.eqn_component,
                       self.profile, boundary)

    # Dispatch routine to call the right boundary-addition gizmo.
    def addToBoundary(self, boundary):
        boundary.addFixedCondition(self)

    # BC is applied through its "call" method, which just sets
    # the value of the appropriate DOF to the passed-in value, plus
    # the offset, and marks the corresponding equation as
    # dependent.
    def __call__(self, node, value):
        fldcomp = self.field.getIndex(self.field_component).integer()
        self.field.setvalue(self.femesh, node, fldcomp, value + self.offset )
        dof = self.field.dof(self.femesh, node, fldcomp)
        dof.fix()

        eqncomp = self.equation.getIndex(self.eqn_component).integer()
        nodalEqn = self.equation.nodaleqn(self.femesh, node, eqncomp)
        nodalEqn.dependent()

    def intersect(self, other): # Other should be another DirichletBC or a FloatBC.
        return self.field == other.field and \
               self.field_component == other.field_component and \
               self.equation == other.equation and \
               self.eqn_component == other.eqn_component

    def display(self):
        return "DirichletBC / %s[%s] / %s[%s]" % (`self.field`,
                                              self.field_component,
                                              `self.equation`,
                                              self.eqn_component)
    
registeredclass.Registration("Dirichlet",
                             BC, DirichletBC,
                             ordering=1,
                             params = bc_field_param_group +
                             bc_equation_param_group + [ 
    parameter.RegisteredParameter(
    'profile', reg=profile.Profile, value=None,
    tip='Profile describing how this condition varies.'),
    bc_boundary_param
    ],
                             tip="Fix the values of a Field along a boundary."
                             ) 


class ForceBC(BC):
    def __init__(self, equation, eqn_component, profile, boundary):
        self.equation = equation
        self.eqn_component = eqn_component

        BC.__init__(self, profile, boundary)

    def copy(self, boundary):
        return ForceBC(self.equation, self.eqn_component,
                       self.profile, boundary)
    #
    # def write(self, file, indent):
    #     file.write(indent+"ForceBC("+problem.inversedict[self.equation]+",")
    #     file.write("%d)" % self.equationComponent)
    #
    def addToBoundary(self, boundary):
        boundary.addForceCondition(self)
    #
    # Applied via "call" method, just creates an applicator if it
    # doesn't already have one, and calls it with the node and
    # equation component, and the value to set.  
    #
    def __call__(self, node, value):
        eqnindex = self.equation.getIndex(self.eqn_component).integer()
        boundarycond.applyForceBC(self.femesh, self.equation,
                                  node, eqnindex, value)
    def display(self):
        return "ForceBC / %s" % `self.equation`



registeredclass.Registration("Generalized Force",
                             BC,
                             ForceBC,
                             ordering=2,
                             params = bc_equation_param_group + [
    parameter.RegisteredParameter(
    'profile', reg=profile.Profile, value=None,
    tip='Profile describing how this condition varies.'),
    bc_boundary_param],
                             tip="Fix the values of a force at all nodes on a boundary.  A force is a divergence of a flux (eg, stress, heat flux).  The condition is applied independently at each node, and is insensitive to node density."
                             )


## Float boundary condition -- similar to a DirichletBC in basic
## structure, but with major differences in the apply function.  What
## 'apply' does is go through the mapping lists passed to the mesh by
## the matrix extraction routine, and set them up so that all the rows
## and columns corresponding to the DOFs and nodalequations of the
## floating boundary condition are remapped to a single row/colum,
## respectively, of the new mapping.

## They also have an unravelling routine which unmaps the DOFs
## after solution.

## The 'profile' function has the same structure but a different
## meaning here -- it's an offset, the true fully-solved value of the
## DOF is the profile value plus another value, such that the sum
## equals the sum arrived at by the solver.


class FloatBC(BC):
    def __init__(self, field, field_component, equation, eqn_component,
                 profile, boundary):
        self.field = field
        self.field_component = field_component
        self.equation = equation
        self.eqn_component = eqn_component
        BC.__init__(self, profile, boundary)
        self.reset()

    def copy(self, boundary):
        return FloatBC(self.field, self.field_component,
                       self.equation, self.eqn_component,
                       self.profile, boundary)
        
    # This function can be called either by init or from the boundary,
    # when it wants to reset stateful boundary conditions.  This
    # provides a handy way of pretend-encapsulating the state variables.
    # FloatBCs are stateful because intersecting combinations of them
    # need to aggregate their degrees of freedom in a sensible way,
    # with consistent profiles and without double-counting.  This is
    # what the traversal process does.
    def reset(self):
        #
        # Below here are the "stateful" data members.  
        self.applicator = None     # C applicator -- "None" means first-call.
        self.dofmappingIndex = -1  # Set either locally or by parents
        self.eqnmappingIndex = -1  #    on the intersection tree.
        self.expandfirst = None    # Detects first call to "expand"
        self.offset = 0.0          # Local profile offset relative to parent.
        self.profileStart = 0.0    # Master start value, set by root.
        self.profileOffset = 0.0   # Cumulative total offset from intersections
        self.active = 1            # 
        self.dupnodes = {}         # Rationalized duplicate set.
        self.nodeshares = {}       # BC-key'd dict, parent : dict of dupnodes.
        self.intersectors = {}     # BC-key'd dict, child : offset
        self.intersected = None    # Intersection state.
        # States are None, "mapped", and "traversed."


        # A FloatBC is active by default, but can be deactivated
        # if it gets promoted to a dirichlet BC.  This happens when
        # the "fix" method is called.
        # The floatappset member helps to encode state information -- the
        # applicator knows its first call, which is why it's important
        # to de-allocate it at the end.
        #   "Activeness" is checked for locally, so that active and
        # inactive floatBC's present the same interface, but inactive
        # ones don't actually do anything.  The only other place
        # activeness is checked for is in the intersection routine.
        #
        # Duplicate nodes can arise from intersection -- the calls
        # simply do nothing at all in that case.  The first node of
        # the parent of an intersection tree is guaranteed not to
        # be duplicate to the parent, so all the order-depdendent
        # stuff still works.


    # Dispatch routine, called from the boundary, calls back to the
    # correct boundary-addition routine.
    def addToBoundary(self, boundary):
        boundary.addFloatCondition(self)
    #
    def __call__(self, node, value):
        # Statefulness is in the floatappset list -- if it's null, this is
        # the first call, you need to set up one for each field component.
        if self.active and not self.dupnodes.has_key(node):
            fldcomp = self.field.getIndex(self.field_component).integer()
            eqncomp = self.equation.getIndex(self.eqn_component).integer()
            value+=self.profileOffset
            if self.applicator:
                # This is a non-first call, just do editmap.
                self.applicator.editmap(self.femesh, value, node,
                                        self.field, fldcomp,
                                        self.equation, eqncomp,
                                        self.dofmappingIndex,
                                        self.eqnmappingIndex,
                                        self.profileStart)
            else:
                # This is the first call to this particular FloatBC.
                self.applicator = boundarycond.FloatBCApp()
                if not self.intersected:
                    # We are either the root of an intersection tree,
                    # or unintersected -- find out what mapping targets to use.
                    self.profileStart = value

                    self.dofmappingIndex = \
                      self.applicator.getdofindex(self.femesh, node,
                                                  self.field, fldcomp)

                    self.eqnmappingIndex = \
                        self.applicator.geteqnindex(
                        self.femesh, node, self.equation, eqncomp)

                    # Broadcast these to the intersecting set.
                    for intersector in self.intersectors.keys():
                        intersector.propagate_(self.dofmappingIndex,
                                               self.eqnmappingIndex,
                                               self.profileStart)
                else:
                    # If we're here, it means this is the first call
                    # to a non-root member of an intersection tree,
                    # so we need to call editmap on the first node,
                    self.applicator.editmap(
                        self.femesh, value, node,
                        self.field, fldcomp,
                        self.equation, eqncomp,
                        self.dofmappingIndex,
                        self.eqnmappingIndex,
                        self.profileStart)

    # This is the function that handles the propagation of parameters
    # through the intersections -- recursively propagates everything.
    def propagate_(self, dofmappingIndex, eqnmappingIndex, profileStart):
        if self.intersected != 'traversed':
            raise ooferror.ErrPyProgrammingeError(
                "Data propagated to incomplete Float.")
        self.dofmappingIndex = dofmappingIndex
        self.eqnmappingIndex = eqnmappingIndex
        self.profileStart = profileStart
        for intersector in self.intersectors.keys():
            intersector.propagate_(dofmappingIndex, eqnmappingIndex,
                                  profileStart)
    #
    def contrib_rhs(self):
        if self.active:
            self.applicator.profile_rhs(self.femesh)
            self.applicator = None  # Null this, anticipating next round.
            self.expandfirst = None

    # After matrix solution, need to fill in the DOF values
    # according to the profile function.  Tuple length-matches not
    # checked, because if they're wrong, we failed in "invoke."
    # This routine breaks if the nodes don't come in the
    # same order as they did in invoke.  (More specifically, the same
    # node must occur first.)

    def expand(self, node, value):
        # print "FloatBC::expand."
        if self.active and not self.dupnodes.has_key(node):
            value += self.profileOffset
            comp = self.field.getIndex(self.field_component).integer()
            if not self.intersected:
                if self.expandfirst==None:
                    self.expandfirst = 1
                else:
                    prev = self.field.value(self.femesh, node, comp)
                    # print "Assigning component ", comp, " of field ", self.field_component, " the value ", value-self.profileStart_prev, " at node ", node
                    self.field.setvalue(self.femesh, node, comp,
                                        value-self.profileStart+prev)
            else: # Intersection case -- profileStart assumed received.
                prev = self.field.value(self.femesh, node, comp)
                self.field.setvalue(self.femesh, node, comp,
                                    value-self.profileStart+prev)
                    

    # Build the equivalent DirichletBC for your field and equation,
    # and deactivate yourself.
    def fix(self, offset):
        self.active = 0
        return DirichletBC( #self.name, self.mesh,
            self.field, self.field_component,
            self.equation, self.eqn_component,
            self.profile, self.boundary, offset=offset, fromFloat=1)
    #
    def intersect(self, other): # Other should be another FloatBC.
        return self.field == other.field and \
               self.field_component == other.field_component and \
               self.equation == other.equation and \
               self.eqn_component == other.eqn_component

    def display(self):
        return "FloatBC / %s[%s] / %s[%s]" % (`self.field`,
                                              self.field_component,
                                              `self.equation`,
                                              self.eqn_component)

registeredclass.Registration("Floating",
                             BC,
                             FloatBC,
                             ordering=3,
                             params= bc_field_param_group +
                             bc_equation_param_group + [
    parameter.RegisteredParameter(
    'profile', reg=profile.Profile, value=None,
    tip='Profile describing how this condition varies.'),
    bc_boundary_param
    ],
                             tip="Fix the values of Fields on the boundary, up to an unspecified offset. Field values on the boundary can change, but Field differences between two points on the boundary will stay fixed."
                             )



########################################################################
# # # # # # # # # # NEUMANN BOUNDARY CONDITION # # # # # # # # # # # # #
########################################################################

from engine.profile import Location
from SWIG.common import coord

# Wrapper function for flux boundary profiles.  Takes a number of
# floating-point values provided from within C++, translates them into
# a Location object, and calls the profile object with the location.
# OPT: If this is too slow, the Location consruction process can be
# moved into the interior of the flux class's BCCallback routines,
# defined in boundarycond.swg.
def flux_locator(profile_obj, posx, posy, distance, fraction):
    pos = coord.Coord(posx, posy)
    loc = Location(pos, s=distance, alpha=fraction)
    return profile_obj(loc)

#
# Actual flux boundary condition.  Simpler than the others, because
# most of the action happens in C++.
class NeumannBC(BC):
    def __init__(self, flux, profile, boundary, normal = None):
        self.flux = flux
        self.normal = normal  
        BC.__init__(self, profile, boundary)

    def copy(self, boundary):
        return NeumannBC(self.flux, self.profile, boundary, normal=self.normal)
    
    def write(self, file, indent):
        file.write(indent+"NeumannBC("+problem.inversedict[self.flux]+")")
    def addToBoundary(self, boundary):
        boundary.addFluxCondition(self)
    #
    # Invoke means to do the integral over the appropriate
    # boundary, and make the contributions to the RHS.
    def __call__(self, edgeset, profile):
        if self.femesh.is_active_flux(self.flux):
            applicator = boundarycond.NeumannBCApp(self.femesh, self.flux, edgeset)
            # For this call, "profile" must be a Python callable.
            applicator.integrate(flux_locator, profile, self.normal)
        else:
            raise ErrSetupError('Attempt to invoke NeumannBC on inactive flux.')
    def display(self):
        return "NeumannBC/%s" % `self.flux`
##    def display(self):
##        return "NeumannBC(flux=%s)" % `self.flux`

# NeumannBC doesn't make sense to apply point-wise, use the ForceBC for
# that.  ForceBC's are specified in terms of equations and components,
# rather than fluxes, because it makes more sense. 


registeredclass.Registration("Neumann",
                             BC,
                             NeumannBC,
                             ordering=4,
                             params = [
    meshparameters.FluxParameter(
    "flux", tip="Flux whose value is specified across this boundary."),

    profile.FluxProfileSetParameter(
    "profile", tip="Profiles describing the prescribed flux."),
    bc_boundary_param,
    parameter.BooleanParameter(
    "normal",
    tip="If true, the boundary condition is given in a local right-handed coordinate system where x is the outward normal to the boundary and y is the tangent direction.  If false, x and y are lab coordinates."
    )],
                             tip="Fix the normal component of the divergence of a Flux (eg, stress, heat flux) along the boundary.  The condition is integrated over element edges, and therefore is sensitive to node density.")

