# -*- python -*-
# $RCSfile: profile.py,v $
# $Revision: 1.44 $
# $Author: reida $
# $Date: 2004/09/16 15:29:46 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import switchboard
from SWIG.common import ooferror
from common import registeredclass
from common import utils
from common.IO import parameter
from common.IO import whoville
from engine import profilefunction
import engine.mesh
import string, types


#######################################################################
#######################################################################
# The profile manager.  "Smart" dictionary that can send signals
# when things are inserted or removed.

class ProfileManager:
    def __init__(self):
        self.data = {}
    def items(self):
        return self.data.items()
    def keys(self):
        return self.data.keys()
    def uniqueName(self, name):
        return utils.uniqueName(name, self.data.keys())
    def rename(self, oldname, newname):
        newname = utils.uniqueName(newname, self.keys(), exclude=oldname)
        if newname != oldname:
            self.data[newname]=self.data[oldname]
            del self.data[oldname]
            self.data[newname].rename(newname)
        switchboard.notify("profile renamed", oldname, newname)
    def replaceProfile(self, name, profile):
        del self.data[name]
        profile.set_name(name) # Inserts itself into this dictionary.
        switchboard.notify("profile edited", name)
    def __setitem__(self, key, value):
        try:
            prof = self.data[key]
        except KeyError:
            self.data[key]=value
            switchboard.notify("profiles changed", key)
        else:
            if not prof.equiv(value):
                raise ooferror.ErrUserError(
                    "Profile name %s collides with existing profile." % key)
            
    def __getitem__(self, key):
        return self.data[key]
    def __delitem__(self, key):
        del self.data[key]
        # Switchboard only notified if no exception occurs.
        switchboard.notify("profiles changed", key)


AllProfiles = ProfileManager()



# Object for aggregating location information for boundaries. 
# All profile __call__ methods take one of these as an argument.

class Location:
    def __init__(self, position,
                 index=None, s=None, alpha=None):
        self.position = position # Expect a Coord or Point object.
        self.index = index
        self.distance = s
        self.fraction = alpha
    def __repr__(self):
        result = "Location("
        attrlist = []
        for s in ['position', 'index', 'distance', 'fraction']:
            if self.__dict__[s] is not None:
                attrlist.append("%s=%s" % (s, `self.__dict__[s]`) )
        result += string.join(attrlist, ", ") + ")"
        return result
                                

class CoreProfile(registeredclass.RegisteredClass):
    registry = []
    
# Special registration class for "Core" profiles -- adds itself to the
# CoreProfile's registration list, and also to the regular Profile
# registration list, via the Registration __init__ call.  All kinds of
# profiles except NamedProfiles should register themselves this way.
class CoreProfileRegistration(registeredclass.Registration):
    def __init__(self, name, regedclass, subclass, ordering,
                 params=[], secret=0, **kwargs):
        registeredclass.Registration.__init__(self, name, regedclass,
                                              subclass, ordering,
                                              params=params, secret=secret,
                                              **kwargs)
        CoreProfile.registry.append(self)
        


# Parent class for all profile objects, including named profiles,
# which are actually containers.
class Profile(registeredclass.RegisteredClass):
    registry = []
    def __init__(self):
        self._name = None
        self.conditions = []

    # Why is object identity used for comparison?
    def __eq__(self, other):
        return id(self)==id(other)

    # Separate equivalence function, to identify functional sameness.
    # This sense of profile equivalence does not depend on which
    # conditions they are applied to, or their names, only on their
    # class and function.  Should be implemeneted in the subclasses.
    def equiv(self, other):
        return 0
            
        
    def is_named(self):
        return 0
    
    # Cloning works via the registration, which means it resets the
    # default parameters.  This is probably harmless.
    def clone(self):
        self.setDefaultParams()
        return self.getRegistration()() # Retrieve and run the registration.


    # "Unpack" routine retrieves the enclosed object from a
    # NamedProfile, but in ordinary profiles, just returns self.
    def unpack(self):
        return self
    
    def set_name(self, name):
        # Once it's named, it should go in the big list of all profiles.
        AllProfiles[name] = self
        self._name = name

    def name(self):
        return self._name

    # Profile is deletable if it is not assigned to any conditions.
    def deletable(self):
        if len(self.conditions)>0:
            return 0 
        return 1
    
    def rename(self, newname):
        self._name = newname
        
    def addCondition(self, condition):
        self.conditions.append(condition)


    def removeCondition(self, condition):
        self.conditions.remove(condition)

    # Descriptor should be short, appears in the boundary condition page
    # widget with the BCs.
    def descriptor(self):
        if self._name:
            return `self._name`
        else:
            return self.description()



# General continuum profiles provide all the location arguments except
# the components of the local normal to their local ProfileFunction.
class ContinuumProfile(Profile):
    def __init__(self, function):
        self.function = function
        Profile.__init__(self)

    # Argument to the __call__ method is a Location object, which
    # promises a .position member, and may provide others.  Set the
    # variables and call the function.
    def __call__(self, location):
        x = location.position[0]
        y = location.position[1]
        i = location.index
        s = location.distance
        alpha = location.fraction
        return self.function(x, y, i, s, alpha, None, None)
    # Output.
    def description(self):
        return `self.function`

    def equiv(self,other):
        if other.__class__==self.__class__:
            if self.function==other.function:
                return 1
        return 0
    
    def __repr__(self):
        return "ContinuumProfile(%s)" % `self.function`

CoreProfileRegistration("Continuum Profile",
                        Profile,
                        ContinuumProfile,
                        ordering=1,
                        params=[
    profilefunction.ProfileFunctionParameter(
    "function", value=profilefunction.ProfileFunction("1.0"),
    tip="A function of x and y.") ])



# Subclasses for common special cases.

class ConstantProfile(ContinuumProfile):
    def __init__(self, value):
        self.value = value 
        Profile.__init__(self)

    # Caller will provide "location", but we can discard it.
    def __call__(self, location):
        return self.value

    def equiv(self,other):
        if self.__class__==other.__class__:
            if self.value==other.value:
                return 1
        return 0
    
    def description(self):
        return `self.value`
    
    def __repr__(self):
        return "ConstantProfile(value=%s)" % `self.value`

CoreProfileRegistration("Constant Profile",
                        Profile,
                        ConstantProfile,
                        ordering=2,
                        params=[
    parameter.FloatParameter("value", 0.0, 
                             tip="Value of this profile.") ])





class LinearProfile(ContinuumProfile):
    def __init__(self, start, end):
        self.start = start
        self.end = end
        Profile.__init__(self)

    def __call__(self, location):
        return location.fraction*(self.end-self.start)+self.start

    def equiv(self, other):
        if self.__class__==other.__class__:
            if self.start==other.start and self.end==other.end:
                return 1
        return 0

    def description(self):
        return "[%s->%s]" % (`self.start`, `self.end`)
    def __repr__(self):
        return "LinearProfile(start=%s, end=%s)" % (`self.start`, `self.end`)
        
CoreProfileRegistration("Linear Profile",
                        Profile,
                        LinearProfile,
                        ordering=3,
                        params=[
    parameter.FloatParameter("start", 1.0, 
                             tip="Start value of this profile."),
    parameter.FloatParameter("end", 1.0, 
                             tip="End value of this profile.")  ])


# At this point, all the "normal" registrations have been created in
# both the Profile and CoreProfile registration lists.  New non-named
# profiles should be done the same way.  The special NamedProfile
# container class is done differently, it *only* exists in the Profile
# registration list.


# Profile container, for putting pre-existing profiles into BC's.
# Profiles pre-exist in the profile dictionary.

# NamedProfiles 
class NamedProfile(Profile):
    def __init__(self, name):
        self.name = name
        self.obj = AllProfiles[self.name]
        self.sb_cb = [
            switchboard.requestCallback("profile renamed", self.rename_profile),
            switchboard.requestCallback("profile edited", self.replace_profile)
            ]

    def is_named(self):
        return 1
    
    def unpack(self):
        return self.obj

    # self.obj has already been renamed.
    def rename_profile(self, oldname, newname):
        if self.name == oldname:
            self.name = newname

    # Called when the profile with this name in AllProfiles has
    # been replaced with a new one, i.e. when profiles are edited.
    def replace_profile(self, name):
        if self.name == name:
            self.obj = AllProfiles[name]
            switchboard.notify("new namedprofile obj", self)
            
    
    def __call__(self, *args):
        return self.obj.__call__(*args)

    def __del__(self):
        map(switchboard.removeCallback, self.sb_cb)
    
    def __repr__(self):
        return "NamedProfile(%s)" % `self.name`


# Profile names have a widget which retrieves the allowed list from
# the enclosing mesh.  Consequently, they need a special parameter
# type, so the widget gets brought up at the right time.
class ProfileNameParameter(parameter.StringParameter):
    def __init__(self, name, value=None, tip=None):
        parameter.StringParameter.__init__(
            self, name, value, tip)


registeredclass.Registration("Named Profile",
                             Profile,
                             NamedProfile,
                             ordering=1000,
                             params = [
    ProfileNameParameter(name="name")])


################################################################

# ProfileSet object -- a Profile work-alike which is a container
# for one or more profiles.  __call__ method should return a
# tuple of results.  Supports a profile-like __call__ method,
# as well as unpack and descriptor, allowing it to sub for an
# actual profile in a NeumannBC.
class FluxProfileSet:
    def __init__(self, profile_list):
        self.data = profile_list
    def __call__(self, location):
        return tuple([x(location) for x in self.data])
    def __len__(self):
        return len(self.data)
    def unpack(self):
        return self
    def get_profiles(self):
        return self.data[:]
    def addCondition(self, condition):
        for p in self.data:
            p.addCondition(condition)
    def removeCondition(self, condition):
        for p in self.data:
            p.removeCondition(condition)
    def descriptor(self):
        return "(%s)" % string.join([x.descriptor() for x in self.data],", ")
    def __repr__(self):
        return "FluxProfileSet(%s)" % `self.data`

utils.OOFdefine('FluxProfileSet', FluxProfileSet)

class FluxProfileSetParameter(parameter.ParameterBase):
    def checker(self, x):
        if type(x)!=types.InstanceType or \
           x.__class__ != FluxProfileSet:
            raise TypeError(
                "Bad type!  Expected FluxProfileSet, got %s." % type(x))
        

