# -*- python -*-
# $RCSfile: skeletonboundary.py,v $
# $Revision: 1.53 $
# $Author: langer $
# $Date: 2004/09/21 19:23:57 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


# This is the SkeletonContext-level object that represents a boundary.
# It has methods allowing the boundary to be propagated throughout the
# context's stack.  It is not, itself, a boundary, i.e. it has no
# edges of its own.  The "boundaryset" dictionary contains weak
# references to all the boundaries in the context which have the name.

# TODO2.0: Propagate new Skeleton boundaries to existing Meshes.

from SWIG.common import ooferror
from common import debug
from engine import skeletonsegment
from engine import boundarymodifier
import weakref

# Special constants for specifying the direction of the map.
MAP_DOWN = 0
MAP_UP = 1


class SkelContextBoundary:
    def __init__(self, context, name, skeleton=None, boundary=None):
        self.context = context
        self.name = name
        self.boundaryset = weakref.WeakKeyDictionary()
        if skeleton==None:
            skeleton = context.getObject()

        if boundary:
            # DeputySkeletons don't have independent boundary
            # information, so all data has to be associated with the
            # sheriff skeleton.
            self.boundaryset[skeleton.sheriffSkeleton()] = boundary

    def size(self, skeleton):
        return self.boundaryset[skeleton.sheriffSkeleton()].size()
        
    def rename(self, newname):
        for skeleton in self.boundaryset.keys():
            skeleton.renameBoundary(self.name, newname)
        self.name = newname
            

class SkelContextEdgeBoundary(SkelContextBoundary):
    modifiertype = boundarymodifier.EdgeBoundaryModifier
    def __init__(self, context, name, skeleton=None, boundary=None):
        SkelContextBoundary.__init__(self, context, name,
                                     skeleton, boundary)

    # Removal function -- if a skeleton is specified, remove the
    # corresponding SkeletonBoundary from the skeleton.  Otherwise, do
    # it for all of them.
    def remove(self, skeleton=None):
        if skeleton:
            self.boundaryset[skeleton].remove()
            skeleton.removeBoundary(self.name)
        else:
            for b in self.boundaryset.values():
                b.remove()
            for k in self.boundaryset.keys():
                k.removeBoundary(self.name)


    # Addition of new segments.  The added segments and nodes are
    # from the current .obj of self.context. 
    def append(self, segments, nodes):
        skelbdy = self.boundaryset[self.context.getObject()]
        skelbdy.append(segments, nodes)


    def delete(self, segments, nodes):
        skelbdy = self.boundaryset[self.context.getObject()]
        skelbdy.delete(segments, nodes)


    # Propagate a boundary up to its parent or down to its child.
    # Caller must ensure that the skeleton passed through as
    # "new_skeleton" in fact contains the appropriately-related
    # segments (parents for MAP_UP, children for MAP_DOWN) with
    # respect to first skeleton.  Caller is also responsible for
    # issuing the "new_boundaries" signal.
    def map(self, skeleton, new_skeleton, direction=MAP_DOWN, exterior=None):
        old_bdy = self.boundaryset[skeleton]
        new_bdy = new_skeleton.makeEdgeBoundary(self.name, exterior=exterior)
        self.boundaryset[new_skeleton] = new_bdy

        # Deduce the corresponding edges in the new mesh.
        old_edges = old_bdy.edges[:]
        while len(old_edges) > 0:
            # Start from the back, so list removal is efficient.
            e = old_edges[-1]
            
            if direction==MAP_DOWN:
                m = e.segment.map()
                map_start = m.source
                map_end = m.target
            else: # direction==MAP_UP
                if len(e.segment.getParents())==0:
                    del old_edges[-1]
                    continue
                m = e.segment.getParents()[0].map()
                map_start = m.target
                map_end = m.source
                

            if len(map_start)==1:
                # One-to-many case -- Sequence and edge-ify.
                new_eset = edgesFromSegs(e, map_end, direction)
                for e in new_eset:
                    new_bdy.addEdge(e)
                del old_edges[-1]
            else:
                # Many-to-one or many-to-many case.
                # First, find all the edges of this bdy in the parent part.
                source_eset = [e]
                for seg in map_start:
                    for edg in seg.edges:
                        if edg in old_edges:
                            if edg not in source_eset:
                                source_eset.append(edg)

                # Remove them from the old-edge list.
                for e in source_eset:
                    del old_edges[-1]
                                
                # Sequence the source edges -- this is the path the
                # original boundary takes through the source part
                # of the map.
                try:
                    (edge_seq, node_seq) = skeletonsegment.segSequence(
                        source_eset)
                except skeletonsegment.SequenceError:
                    continue # Re-enter the while loop.
                
                # Direct the resulting edge-sequence correctly.  Edges
                # always return their nodes in order.
                if node_seq[0]!=edge_seq[0].get_nodes()[0]:
                    edge_seq.reverse()
                    node_seq.reverse()
                    
                # Extract the subset of the target that has some
                # relation to the boundary part of the parent.
                sub_target = []
                for e in edge_seq:
                    if direction==MAP_DOWN:
                        for c in e.segment.getChildren():
                            if c not in sub_target:
                                sub_target.append(c)
                    else: # direction==MAP_UP:
                        for p in e.segment.getParents():
                            if p not in sub_target:
                                sub_target.append(p)

                # Exterior check -- for the exterior case, the only
                # valid target segments are those which have exactly one
                # element.
                if exterior:
                    narrow_target = []
                    for s in sub_target:
                        if s.nElements()==1:
                            narrow_target.append(s)
                    sub_target = narrow_target
                        
                # Find the counterparts of the endpoints of the path, if
                # they exist and are unique, otherwise fail.
                if direction==MAP_DOWN:
                    start_shadow = node_seq[0].getChildren()
                    end_shadow = node_seq[-1].getChildren()
                else:
                    start_shadow = node_seq[0].getParents()
                    end_shadow = node_seq[-1].getParents()

                if len(start_shadow)==1:
                    target_start = start_shadow[0]
                else:
                    continue


                if len(end_shadow)==1:
                    target_end = end_shadow[0]
                else:
                    continue

                # If the points are not distinct, fail.
                if target_start == target_end:
                    continue
                
                # Find the paths in the target subset segment set from
                # the start node to the end node.  The returned path
                # is sequenced.
                path_set = skeletonsegment.segPath(
                    target_start, target_end, sub_target)

                # If the path is not unique, fail.
                if len(path_set)!=1:
                    continue

                # Success!  We have a single unique sequenced set of
                # segments from child_start to child_end.  Convert
                # them to edges and add them to the boundary.
                n1 = target_start
                for seg in path_set[0]:
                    nodes = seg.get_nodes()
                    if nodes[0]==n1:
                        e = skeletonsegment.SkeletonEdge(seg, direction=1)
                        n2 = nodes[1]
                    else:
                        e = skeletonsegment.SkeletonEdge(seg, direction=-1)
                        n2 = nodes[0]
                    new_bdy.addEdge(e)
                    n1 = n2 # Reset n1 for next iteration.

class ExteriorSkelContextEdgeBoundary(SkelContextEdgeBoundary):
    # Exterior version needs to restrict itself to new segments
    # which have only one element.  Set the flag in the base
    # class and call it.
    def map(self, skeleton, new_skeleton, direction=MAP_DOWN):
        SkelContextEdgeBoundary.map(self, skeleton, new_skeleton,
                                    direction, exterior=1)


            
class SkelContextPointBoundary(SkelContextBoundary):
    modifiertype = boundarymodifier.PointBoundaryModifier
    def __init__(self, context, name, skeleton=None, boundary=None):
        SkelContextBoundary.__init__(self, context, name,
                                     skeleton, boundary)

    # Removal is simpler for point boundaries, just remove it from the
    # specified skeleton, or all of them if none is specified.
    def remove(self, skeleton=None):
        if skeleton:
            skeleton.removeBoundary(self.name)
        else:
            for k in self.boundaryset.keys():
                k.removeBoundary(self.name)


    def append(self, nodes):
        skelbdy = self.boundaryset[self.context.getObject()]
        skelbdy.append(nodes)

    def delete(self, nodes):
        skelbdy = self.boundaryset[self.context.getObject()]
        skelbdy.delete(nodes)
            
    # Point boundary propagation is much easier.
    def map(self, skeleton, new_skeleton, direction=MAP_DOWN, exterior=None):
        old_bdy = self.boundaryset[skeleton]
        new_bdy = new_skeleton.makePointBoundary(self.name, exterior=exterior)
        self.boundaryset[new_skeleton] = new_bdy

        target_list = []
        for n in old_bdy.nodes:
            if direction==MAP_DOWN:
                sub_nodes = n.getChildren()
            else:
                sub_nodes = n.getParents()

            for c in sub_nodes:
                if exterior:
                    if (c.movable_x()==n.movable_x()) and \
                       (c.movable_y()==n.movable_y()):
                        target_list.append(c)
                else:
                    target_list.append(c)

        new_node_list = []
        for n in target_list:
            valid = 1
            if direction==MAP_DOWN:
                sources = n.getParents()
            else:
                sources = n.getChildren()
                
            for c in sources:
                if not old_bdy.has_node(c):
                    valid = 0

            if valid==1:
                new_node_list.append(n) # Only reached if all c's present.

        for n in new_node_list:
            new_bdy.addNode(n)


class ExteriorSkelContextPointBoundary(SkelContextPointBoundary):
    def map(self, skeleton, new_skeleton, direction=MAP_DOWN):
        SkelContextPointBoundary.map(self, skeleton, new_skeleton,
                                     direction, exterior=1)

        
class SkeletonEdgeBoundary: # corresponds to a realskeleton's EdgeBoundary
    def __init__(self, name):
        self._name = name
        self.edges = []
    def size(self):
        return len(self.edges)
    def addEdge(self, edge):
        self.edges.append(edge)

    # This routine removes all the edges from their associated segments.
    # It should be called at boundary-deletion time. 
    def remove(self):
        for e in self.edges:
            e.remove()


    # Function for ensuring that the "self.edges" list is in the
    # correct order, and proceeds from start to finish.  Should work
    # in such a way that it's very fast if the boundary is already
    # sequenced.
    def sequence(self):
        if len(self.edges)==0 or len(self.edges)==1:
            return
        for i in range(len(self.edges)-1):
            if self.edges[i].get_nodes()[1]!=self.edges[i+1].get_nodes()[0]:
                break
        else:
            return

        # Edges are not correctly sequenced.  Do it here, remembering
        # that nodes within edges are already ordered.
        ndict = {}
        startnodes = []
        endnodes = []
        for e in self.edges:
            startnode = e.get_nodes()[0]
            try:
                ndict[startnode]
            except KeyError:
                ndict[startnode]=e
                startnodes.append(startnode)
                endnodes.append(e.get_nodes()[1])
            else:
                raise skeletonsegment.SequenceError(
                    "Cannot sequence boundary %s" % `self._name`)

        # At this point, you have a list of start nodes, a list of end
        # nodes, and a dictionary of edges keyed by start node.

        # Topology check -- eliminate start nodes with corresponding
        # end nodes, and look at the sizes of the remainder lists --
        # they must be equal, and if they're both 1, it's a line, if
        # they're 0, it's a loop, otherwise it's an error.
        unmatchedendnodes = []
        for e in endnodes:
            try:
                startnodes.remove(e)
            except ValueError:
                unmatchedendnodes.append(e)

        if len(unmatchedendnodes)==1 and len(startnodes)==1:
            # Straightforward case.
            sequence_node = startnodes[0]
        elif len(unmatchedendnodes)==0 and len(startnodes)==0:
            # Loop case.  Doesn't matter where you start.
            sequence_node = self.edges[0].get_nodes()[0]
        else:
            raise skeletonsegment.SequenceError(
                "Topology problem with boundary %s, unable to sequence." % `self._name`)

        # Now, finally, we have a starting node.  Chain the edges
        # together from this node.  Proceed until dictionary retrieval
        # on the next node fails.

        # This really wants to be a "do...while" loop, but Python
        # doesn't offer this construction.
        newedgelist = []
        while 1:
            try:
                edge = ndict[sequence_node]
            except KeyError:
                break
            else:
                newedgelist.append(edge)
                sequence_node = edge.get_nodes()[1]

        if len(newedgelist)==len(self.edges):
            self.edges = newedgelist
        else:
            # This can occur if the boundary is made up of disjoint loops.
            raise skeletonsegment.SequenceError(
                "Sequenced boundary %s did not use all edges." % `self._name`)
        
        

    # Double dispatch, returns the right type of context object.
    def makeContextBoundary(self, context, name, skeleton):
        return SkelContextEdgeBoundary(context, name, skeleton, self)

    # Append the indicated segments to this boundary, if possible.
    # The segments are sequenced, but may be in the wrong order.  They
    # may intersect the boundary anywhere or nowhere -- check all
    # cases.
    def append(self, segments, nodes):

        reverse = 0

        if len(self.edges)>0:
            
            # Build a node-dict. 
            ndict = {}
            for e in self.edges:
                for n in e.get_nodes():
                    try:
                        ndict[n].append(e)
                    except KeyError:
                        ndict[n] = [e]

            # "startok" gets set to one if nodes[0] is present in
            # ndict corresponding to precisely one segment.
            startok = 0
            try:
                elist = ndict[nodes[0]]
            except KeyError:
                pass
            else:
                if len(elist)==1:
                    startok = 1

            # "endok" gets set to one if nodes[-1] is present in
            # ndict corresponding to precisely one segment.
            endok = 0
            try:
                elist = ndict[nodes[-1]]
            except KeyError:
                pass
            else:
                if len(elist)==1:
                    endok = 1

            # "middlehit" gets set to one if any of the non-endpoint
            # nodes of the passed-in set is found anywhere in the boundary.
            middlehit = 0
            for n in nodes[1:-1]:
                try:
                    ndict[n]
                except KeyError:
                    pass
                else:
                    middlehit = 1
                    break   

            if middlehit or not (startok or endok):
                raise ooferror.ErrSetupError(
                    "Cannot append indicated segments to boundary.")

            if endok and not startok:
                segments.reverse()
                nodes.reverse()
            
            if ndict[nodes[0]][0].get_nodes()[0]==nodes[0]:
                reverse = 1

        # End of "if len(self.edges)>0".  The following code is
        # executed directly (with reverse=0) if the boundary is initially
        # empty.

        current_node = nodes[0]
        for (s,n) in zip(segments, nodes[1:]):
            if s.nodes()[0]==current_node:
                if reverse:
                    newedge = skeletonsegment.SkeletonEdge(s,-1)
                else:
                    newedge = skeletonsegment.SkeletonEdge(s,1)
            else:
                if reverse:
                    newedge = skeletonsegment.SkeletonEdge(s,1)
                else:
                    newedge = skeletonsegment.SkeletonEdge(s,-1)
                    
            self.addEdge(newedge)
            current_node = n

    # Make sure all the segments are present in the boundary, and that
    # at least one endpoint is shared.  Then remove the corresponding
    # edges.
    def delete(self, segments, nodes):

        # Build a node-dict. 
        ndict = {}
        for e in self.edges:
            for n in e.get_nodes():
                try:
                    ndict[n].append(e)
                except KeyError:
                    ndict[n] = [e]

        startok = 0
        try:
            elist = ndict[nodes[0]]
        except KeyError:
            pass
        else:
            if len(elist)==1:
                startok = 1

        endok = 0
        try:
            elist = ndict[nodes[-1]]
        except KeyError:
            pass
        else:
            if len(elist)==1:
                endok = 1

        if startok or endok:
            segset = segments[:]
            for e in self.edges:
                if e.segment in segset:
                    segset.remove(e.segment)
            if len(segset)!=0:
                raise ooferror.ErrSetupError(
                    "Not all segments present in the boundary.")

            # Now actually remove the edges.
            for e in self.edges[:]:
                if e.segment in segments:
                    e.remove() # Removes from the segment's list.
                    self.edges.remove(e)
                    

        else:
            raise ooferror.ErrSetupError(
                "Segment set does not connect with boundary.")
            
    # temp __repr__ for skeleton save
    def __repr__(self):
        return "SkeletonEdgeBoundary()"
    
    def get_name(self):
        return self._name
class ExteriorSkeletonEdgeBoundary(SkeletonEdgeBoundary):
    def addEdge(self, edge):
        if edge.segment.nElements() != 1:
            raise ooferror.ErrPyProgrammingError(
                "Attempt to insert interior segment in exterior boundary.")
        self.edges.append(edge)
    def makeContextBoundary(self, context, name, skeleton):
        return ExteriorSkelContextEdgeBoundary(context, name, skeleton, self)
    # temp __repr__ for skeleton save
    def __repr__(self):
        return "ExteriorSkeletonEdgeBoundary()"

class SkeletonPointBoundary: # corresponds to a realskeleton's NodeSet
    def __init__(self, name):
        self._name = name
        self.nodes = []
    def size(self):
        return len(self.nodes)
    def addNode(self, node):
        self.nodes.append(node)
    
    def append(self, nodes):
        for n in nodes:
            if n not in self.nodes:
                self.addNode(n)

    def delete(self, nodes):
        for n in nodes:
            if n in self.nodes:
                self.nodes.remove(n)
                
    def has_node(self, node):
        return node in self.nodes
    # Double dispatch, to get correct type of context object.
    def makeContextBoundary(self, context, name, skeleton):
        return SkelContextPointBoundary(context, name, skeleton, self)
    # temp __repr__ for skeleton save
    def __repr__(self):
        return "SkeletonPointBoundary()"
    
    def get_name(self):
        return self._name
    
class ExteriorSkeletonPointBoundary(SkeletonPointBoundary):
    def makeContextBoundary(self, context, name, skeleton):
        return ExteriorSkelContextPointBoundary(context, name, skeleton, self)
    # temp __repr__ for skeleton save
    def __repr__(self):
        return "ExteriorSkeletonPointBoundary()"



# This function returns a list of edges with the properties that:
#  - The target list edges trace a path from the edge's start node
#                      to the edge's stop node.
#  - The target edge's segments are related (parents if
#                      direction is up, children if direction
#                      is down) to the edge's segment.
#  - The new edges are directed correctly from start to finish.
def edgesFromSegs(edge, target_segs, direction):
    if len(target_segs)==0:
        return []
    
    # If the target segments are forked or otherwise complex,
    # return an empty list.
    try:
        (seg_list, node_list) = skeletonsegment.segSequence( target_segs )
    except skeletonsegment.SequenceError:
        return []
    
    ordered_nodes = edge.get_nodes()

    if direction==MAP_UP:
        target_start = ordered_nodes[0].getParents()
    else: # direction==MAP_DOWN
        target_start = ordered_nodes[0].getChildren()
        
    if node_list[0] in target_start:
        pass
    elif node_list[-1] in target_start:
        seg_list.reverse()
        node_list.reverse()
    else:
        raise ooferror.ErrPyProgrammingError(
            "Malformed segment sequence -- node counterpart not found.")
    
    target_list = []
    for (s, n)  in map(None, seg_list, node_list[:-1]):
        if s.nodes()[0]==n:
            target_list.append(skeletonsegment.SkeletonEdge(s,direction=1))
        else:
            target_list.append(skeletonsegment.SkeletonEdge(s,direction=-1))
            
    return target_list

from common import utils
utils.OOFdefine('SkeletonEdgeBoundary', SkeletonEdgeBoundary)
utils.OOFdefine('ExteriorSkeletonEdgeBoundary', ExteriorSkeletonEdgeBoundary)
utils.OOFdefine('SkeletonPointBoundary', SkeletonPointBoundary)
utils.OOFdefine('ExteriorSkeletonPointBoundary', ExteriorSkeletonPointBoundary)
